@@ -1,5 +1,8 @@
 /*
- * Copyright 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,28 +13,29 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: chsh.c,v 1.39 2006/02/21 22:44:35 kloczek Exp $"
+#ident "$Id: chsh.c 1980 2008-04-27 00:40:09Z nekral-guest $"
 
 #include <fcntl.h>
+#include <getopt.h>
 #include <pwd.h>
 #include <signal.h>
 #include <stdio.h>
@@ -40,9 +44,6 @@
 #include <selinux/selinux.h>
 #include <selinux/av_permissions.h>
 #endif
-#ifdef HAVE_SHADOW_H
-#include <shadow.h>
-#endif
 #include "defines.h"
 #include "exitcodes.h"
 #include "getdef.h"
@@ -62,6 +63,11 @@
 static char *Prog;		/* Program name */
 static int amroot;		/* Real UID is root */
 static char loginsh[BUFSIZ];	/* Name of new login shell */
+/* command line options */
+static int sflg = 0;		/* -s - set shell from command line  */
+#ifdef USE_PAM
+static pam_handle_t *pamh = NULL;
+#endif
 
 /* external identifiers */
 
@@ -70,13 +76,21 @@ static void usage (void);
 static void new_fields (void);
 static int check_shell (const char *);
 static int restricted_shell (const char *);
+static void process_flags (int argc, char **argv);
+static void check_perms (const struct passwd *pw);
+static void update_shell (const char *user, char *loginsh);
 
 /*
  * usage - print command line syntax and exit
  */
 static void usage (void)
 {
-	fprintf (stderr, _("Usage: %s [-s shell] [name]\n"), Prog);
+	fputs (_("Usage: chsh [options] [LOGIN]\n"
+	         "\n"
+	         "Options:\n"
+	         "  -h, --help                    display this help message and exit\n"
+	         "  -s, --shell SHELL             new login shell for the user account\n"
+	         "\n"), stderr);
 	exit (E_USAGE);
 }
 
@@ -88,7 +102,7 @@ static void usage (void)
  */
 static void new_fields (void)
 {
-	printf (_("Enter the new value, or press ENTER for the default\n"));
+	puts (_("Enter the new value, or press ENTER for the default"));
 	change_field (loginsh, sizeof loginsh, _("Login Shell"));
 }
 
@@ -131,8 +145,9 @@ static int check_shell (const char *sh)
 #ifdef HAVE_GETUSERSHELL
 	setusershell ();
 	while ((cp = getusershell ())) {
-		if (*cp == '#')
+		if (*cp == '#') {
 			continue;
+		}
 
 		if (strcmp (cp, sh) == 0) {
 			found = 1;
@@ -141,15 +156,20 @@ static int check_shell (const char *sh)
 	}
 	endusershell ();
 #else
-	if ((fp = fopen (SHELLS_FILE, "r")) == (FILE *) 0)
+	fp = fopen (SHELLS_FILE, "r");
+	if (NULL == fp) {
 		return 0;
+	}
 
 	while (fgets (buf, sizeof (buf), fp)) {
-		if ((cp = strrchr (buf, '\n')))
+		cp = strrchr (buf, '\n');
+		if (NULL != cp) {
 			*cp = '\0';
+		}
 
-		if (buf[0] == '#')
+		if (buf[0] == '#') {
 			continue;
+		}
 
 		if (strcmp (buf, sh) == 0) {
 			found = 1;
@@ -162,51 +182,27 @@ static int check_shell (const char *sh)
 }
 
 /*
- * chsh - this command controls changes to the user's shell
+ *  * process_flags - parse the command line options
  *
- *	The only supported option is -s which permits the the login shell to
- *	be set from the command line.
+ *	It will not return if an error is encountered.
  */
-int main (int argc, char **argv)
+static void process_flags (int argc, char **argv)
 {
-	char *user;		/* User name                         */
-	int flag;		/* Current command line flag         */
-	int sflg = 0;		/* -s - set shell from command line  */
-	const struct passwd *pw;	/* Password entry from /etc/passwd   */
-	struct passwd pwent;	/* New password entry                */
+	int option_index = 0;
+	int c;
+	static struct option long_options[] = {
+		{"help", no_argument, NULL, 'h'},
+		{"shell", required_argument, NULL, 's'},
+		{NULL, 0, NULL, '\0'}
+	};
 
-#ifdef USE_PAM
-	pam_handle_t *pamh = NULL;
-	struct passwd *pampw;
-	int retval;
-#endif
-
-	sanitize_env ();
-
-	setlocale (LC_ALL, "");
-	bindtextdomain (PACKAGE, LOCALEDIR);
-	textdomain (PACKAGE);
-
-	/*
-	 * This command behaves different for root and non-root users.
-	 */
-	amroot = getuid () == 0;
-
-	/*
-	 * Get the program name. The program name is used as a prefix to
-	 * most error messages.
-	 */
-	Prog = Basename (argv[0]);
-
-	OPENLOG ("chsh");
-
-	/*
-	 * There is only one option, but use getopt() anyway to
-	 * keep things consistent.
-	 */
-
-	while ((flag = getopt (argc, argv, "s:")) != EOF) {
-		switch (flag) {
+	while ((c =
+		getopt_long (argc, argv, "hs:", long_options,
+		             &option_index)) != -1) {
+		switch (c) {
+		case 'h':
+			usage ();
+			break;
 		case 's':
 			sflg++;
 			STRFCPY (loginsh, optarg);
@@ -220,54 +216,26 @@ int main (int argc, char **argv)
 	 * There should be only one remaining argument at most and it should
 	 * be the user's name.
 	 */
-	if (argc > optind + 1)
+	if (argc > (optind + 1)) {
 		usage ();
-
-	/*
-	 * Get the name of the user to check. It is either the command line
-	 * name, or the name getlogin() returns.
-	 */
-	if (optind < argc) {
-		user = argv[optind];
-		pw = getpwnam (user);
-		if (!pw) {
-			fprintf (stderr,
-				 _("%s: unknown user %s\n"), Prog, user);
-			exit (1);
-		}
-	} else {
-		pw = get_my_pwent ();
-		if (!pw) {
-			fprintf (stderr,
-				 _
-				 ("%s: Cannot determine your user name.\n"),
-				 Prog);
-			exit (1);
-		}
-		user = xstrdup (pw->pw_name);
 	}
+}
 
-#ifdef	USE_NIS
-	/*
-	 * Now we make sure this is a LOCAL password entry for this user ...
-	 */
-	if (__ispwNIS ()) {
-		char *nis_domain;
-		char *nis_master;
-
-		fprintf (stderr,
-			 _("%s: cannot change user `%s' on NIS client.\n"),
-			 Prog, user);
-
-		if (!yp_get_default_domain (&nis_domain) &&
-		    !yp_master (nis_domain, "passwd.byname", &nis_master)) {
-			fprintf (stderr,
-				 _
-				 ("%s: `%s' is the NIS master for this client.\n"),
-				 Prog, nis_master);
-		}
-		exit (1);
-	}
+/*
+ * check_perms - check if the caller is allowed to add a group
+ *
+ *	Non-root users are only allowed to change their shell, if their current
+ *	shell is not a restricted shell.
+ *
+ *	Non-root users must be authenticated.
+ *
+ *	It will not return if the user is not allowed.
+ */
+static void check_perms (const struct passwd *pw)
+{
+#ifdef USE_PAM
+	int retval;
+	struct passwd *pampw;
 #endif
 
 	/*
@@ -275,10 +243,11 @@ int main (int argc, char **argv)
 	 * UID of the user matches the current real UID.
 	 */
 	if (!amroot && pw->pw_uid != getuid ()) {
-		SYSLOG ((LOG_WARN, "can't change shell for `%s'", user));
+		SYSLOG ((LOG_WARN, "can't change shell for `%s'", pw->pw_name));
 		closelog ();
 		fprintf (stderr,
-			 _("You may not change the shell for %s.\n"), user);
+		         _("You may not change the shell for %s.\n"),
+		         pw->pw_name);
 		exit (1);
 	}
 
@@ -287,10 +256,11 @@ int main (int argc, char **argv)
 	 * is not a restricted one.
 	 */
 	if (!amroot && restricted_shell (pw->pw_shell)) {
-		SYSLOG ((LOG_WARN, "can't change shell for `%s'", user));
+		SYSLOG ((LOG_WARN, "can't change shell for `%s'", pw->pw_name));
 		closelog ();
 		fprintf (stderr,
-			 _("You may not change the shell for %s.\n"), user);
+		         _("You may not change the shell for %s.\n"),
+		         pw->pw_name);
 		exit (1);
 	}
 #ifdef WITH_SELINUX
@@ -301,10 +271,11 @@ int main (int argc, char **argv)
 	if ((pw->pw_uid != getuid ())
 	    && (is_selinux_enabled () > 0)
 	    && (selinux_check_passwd_access (PASSWD__CHSH) != 0)) {
-		SYSLOG ((LOG_WARN, "can't change shell for `%s'", user));
+		SYSLOG ((LOG_WARN, "can't change shell for `%s'", pw->pw_name));
 		closelog ();
 		fprintf (stderr,
-			 _("You may not change the shell for %s.\n"), user);
+		         _("You may not change the shell for %s.\n"),
+		         pw->pw_name);
 		exit (1);
 	}
 #endif
@@ -316,13 +287,14 @@ int main (int argc, char **argv)
 	 * before any changes can be made. Idea from util-linux
 	 * chfn/chsh.  --marekm
 	 */
-	if (!amroot && getdef_bool ("CHSH_AUTH"))
+	if (!amroot && getdef_bool ("CHSH_AUTH")) {
 		passwd_check (pw->pw_name, pw->pw_passwd, "chsh");
+        }
 
 #else				/* !USE_PAM */
 	retval = PAM_SUCCESS;
 
-	pampw = getpwuid (getuid ());
+	pampw = getpwuid (getuid ()); /* local, no need for xgetpwuid */
 	if (pampw == NULL) {
 		retval = PAM_USER_UNKNOWN;
 	}
@@ -350,13 +322,193 @@ int main (int argc, char **argv)
 		exit (E_NOPERM);
 	}
 #endif				/* USE_PAM */
+}
+
+/*
+ * update_shell - update the user's shell in the passwd database
+ *
+ *	Commit the user's entry after changing her shell field.
+ *
+ *	It will not return in case of error.
+ */
+static void update_shell (const char *user, char *newshell)
+{
+	const struct passwd *pw;	/* Password entry from /etc/passwd   */
+	struct passwd pwent;		/* New password entry                */
+
+	/*
+	 * Before going any further, raise the ulimit to prevent
+	 * colliding into a lowered ulimit, and set the real UID
+	 * to root to protect against unexpected signals. Any
+	 * keyboard signals are set to be ignored.
+	 */
+	if (setuid (0) != 0) {
+		SYSLOG ((LOG_ERR, "can't setuid(0)"));
+		closelog ();
+		fputs (_("Cannot change ID to root.\n"), stderr);
+		exit (1);
+	}
+	pwd_init ();
+
+	/*
+	 * The passwd entry is now ready to be committed back to
+	 * the password file. Get a lock on the file and open it.
+	 */
+	if (pw_lock () == 0) {
+		SYSLOG ((LOG_WARN, "can't lock /etc/passwd"));
+		closelog ();
+		fputs (_("Cannot lock the password file; try again later.\n"),
+		       stderr);
+		exit (1);
+	}
+	if (pw_open (O_RDWR) == 0) {
+		SYSLOG ((LOG_ERR, "can't open /etc/passwd"));
+		closelog ();
+		fputs (_("Cannot open the password file.\n"), stderr);
+		pw_unlock ();
+		exit (1);
+	}
+
+	/*
+	 * Get the entry to update using pw_locate() - we want the real
+	 * one from /etc/passwd, not the one from getpwnam() which could
+	 * contain the shadow password if (despite the warnings) someone
+	 * enables AUTOSHADOW (or SHADOW_COMPAT in libc).  --marekm
+	 */
+	pw = pw_locate (user);
+	if (NULL == pw) {
+		pw_unlock ();
+		fprintf (stderr,
+		         _("%s: %s not found in /etc/passwd\n"), Prog, user);
+		exit (1);
+	}
+
+	/*
+	 * Make a copy of the entry, then change the shell field. The other
+	 * fields remain unchanged.
+	 */
+	pwent = *pw;
+	pwent.pw_shell = newshell;
+
+	/*
+	 * Update the passwd file entry. If there is a DBM file, update
+	 * that entry as well.
+	 */
+	if (pw_update (&pwent) == 0) {
+		SYSLOG ((LOG_ERR, "error updating passwd entry"));
+		closelog ();
+		fputs (_("Error updating the password entry.\n"), stderr);
+		pw_unlock ();
+		exit (1);
+	}
+
+	/*
+	 * Changes have all been made, so commit them and unlock the file.
+	 */
+	if (pw_close () == 0) {
+		SYSLOG ((LOG_ERR, "can't rewrite /etc/passwd"));
+		closelog ();
+		fputs (_("Cannot commit password file changes.\n"), stderr);
+		pw_unlock ();
+		exit (1);
+	}
+	if (pw_unlock () == 0) {
+		SYSLOG ((LOG_ERR, "can't unlock /etc/passwd"));
+		closelog ();
+		fputs (_("Cannot unlock the password file.\n"), stderr);
+		exit (1);
+	}
+}
+
+/*
+ * chsh - this command controls changes to the user's shell
+ *
+ *	The only supported option is -s which permits the the login shell to
+ *	be set from the command line.
+ */
+int main (int argc, char **argv)
+{
+	char *user;		/* User name                         */
+	const struct passwd *pw;	/* Password entry from /etc/passwd   */
+
+	sanitize_env ();
+
+	setlocale (LC_ALL, "");
+	bindtextdomain (PACKAGE, LOCALEDIR);
+	textdomain (PACKAGE);
+
+	/*
+	 * This command behaves different for root and non-root users.
+	 */
+	amroot = getuid () == 0;
+
+	/*
+	 * Get the program name. The program name is used as a prefix to
+	 * most error messages.
+	 */
+	Prog = Basename (argv[0]);
+
+	OPENLOG ("chsh");
+
+	/* parse the command line options */
+	process_flags (argc, argv);
+
+	/*
+	 * Get the name of the user to check. It is either the command line
+	 * name, or the name getlogin() returns.
+	 */
+	if (optind < argc) {
+		user = argv[optind];
+		pw = xgetpwnam (user);
+		if (!pw) {
+			fprintf (stderr,
+			         _("%s: unknown user %s\n"), Prog, user);
+			exit (1);
+		}
+	} else {
+		pw = get_my_pwent ();
+		if (!pw) {
+			fprintf (stderr,
+			         _
+			         ("%s: Cannot determine your user name.\n"),
+			         Prog);
+			exit (1);
+		}
+		user = xstrdup (pw->pw_name);
+	}
+
+#ifdef	USE_NIS
+	/*
+	 * Now we make sure this is a LOCAL password entry for this user ...
+	 */
+	if (__ispwNIS ()) {
+		char *nis_domain;
+		char *nis_master;
+
+		fprintf (stderr,
+		         _("%s: cannot change user '%s' on NIS client.\n"),
+		         Prog, user);
+
+		if (!yp_get_default_domain (&nis_domain) &&
+		    !yp_master (nis_domain, "passwd.byname", &nis_master)) {
+			fprintf (stderr,
+			         _
+			         ("%s: '%s' is the NIS master for this client.\n"),
+			         Prog, nis_master);
+		}
+		exit (1);
+	}
+#endif
+
+	check_perms (pw);
 
 	/*
 	 * Now get the login shell. Either get it from the password
 	 * file, or use the value from the command line.
 	 */
-	if (!sflg)
+	if (!sflg) {
 		STRFCPY (loginsh, pw->pw_shell);
+	}
 
 	/*
 	 * If the login shell was not set on the command line, let the user
@@ -384,98 +536,17 @@ int main (int argc, char **argv)
 		exit (1);
 	}
 
-	/*
-	 * Before going any further, raise the ulimit to prevent
-	 * colliding into a lowered ulimit, and set the real UID
-	 * to root to protect against unexpected signals. Any
-	 * keyboard signals are set to be ignored.
-	 */
-	if (setuid (0)) {
-		SYSLOG ((LOG_ERR, "can't setuid(0)"));
-		closelog ();
-		fprintf (stderr, _("Cannot change ID to root.\n"));
-		exit (1);
-	}
-	pwd_init ();
+	update_shell (user, loginsh);
 
-	/*
-	 * The passwd entry is now ready to be committed back to
-	 * the password file. Get a lock on the file and open it.
-	 */
-	if (!pw_lock ()) {
-		SYSLOG ((LOG_WARN, "can't lock /etc/passwd"));
-		closelog ();
-		fprintf (stderr,
-			 _
-			 ("Cannot lock the password file; try again later.\n"));
-		exit (1);
-	}
-	if (!pw_open (O_RDWR)) {
-		SYSLOG ((LOG_ERR, "can't open /etc/passwd"));
-		closelog ();
-		fprintf (stderr, _("Cannot open the password file.\n"));
-		pw_unlock ();
-		exit (1);
-	}
-
-	/*
-	 * Get the entry to update using pw_locate() - we want the real
-	 * one from /etc/passwd, not the one from getpwnam() which could
-	 * contain the shadow password if (despite the warnings) someone
-	 * enables AUTOSHADOW (or SHADOW_COMPAT in libc).  --marekm
-	 */
-	pw = pw_locate (user);
-	if (!pw) {
-		pw_unlock ();
-		fprintf (stderr,
-			 _("%s: %s not found in /etc/passwd\n"), Prog, user);
-		exit (1);
-	}
-
-	/*
-	 * Make a copy of the entry, then change the shell field. The other
-	 * fields remain unchanged.
-	 */
-	pwent = *pw;
-	pwent.pw_shell = loginsh;
-
-	/*
-	 * Update the passwd file entry. If there is a DBM file, update
-	 * that entry as well.
-	 */
-	if (!pw_update (&pwent)) {
-		SYSLOG ((LOG_ERR, "error updating passwd entry"));
-		closelog ();
-		fprintf (stderr, _("Error updating the password entry.\n"));
-		pw_unlock ();
-		exit (1);
-	}
-
-	/*
-	 * Changes have all been made, so commit them and unlock the file.
-	 */
-	if (!pw_close ()) {
-		SYSLOG ((LOG_ERR, "can't rewrite /etc/passwd"));
-		closelog ();
-		fprintf (stderr, _("Cannot commit password file changes.\n"));
-		pw_unlock ();
-		exit (1);
-	}
-	if (!pw_unlock ()) {
-		SYSLOG ((LOG_ERR, "can't unlock /etc/passwd"));
-		closelog ();
-		fprintf (stderr, _("Cannot unlock the password file.\n"));
-		exit (1);
-	}
 	SYSLOG ((LOG_INFO, "changed user `%s' shell to `%s'", user, loginsh));
 
 	nscd_flush_cache ("passwd");
 
 #ifdef USE_PAM
-	if (retval == PAM_SUCCESS)
-		pam_end (pamh, PAM_SUCCESS);
+	pam_end (pamh, PAM_SUCCESS);
 #endif				/* USE_PAM */
 
 	closelog ();
 	exit (E_SUCCESS);
 }
+
