@@ -1,5 +1,8 @@
 /*
- * Copyright 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1989 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2001 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,26 +13,26 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: passwd.c,v 1.57 2006/02/21 22:44:35 kloczek Exp $"
+#ident "$Id: passwd.c 2003 2008-05-24 13:08:58Z nekral-guest $"
 
 #include <errno.h>
 #include <fcntl.h>
@@ -40,7 +43,9 @@
 #include <sys/types.h>
 #ifdef WITH_SELINUX
 #include <selinux/selinux.h>
+#include <selinux/flask.h>
 #include <selinux/av_permissions.h>
+#include <selinux/context.h>
 #endif
 #include <time.h>
 #include "defines.h"
@@ -50,15 +55,6 @@
 #include "pwauth.h"
 #include "pwio.h"
 #include "shadowio.h"
-#ifndef GPASSWD_PROGRAM
-#define GPASSWD_PROGRAM "gpasswd"
-#endif
-#ifndef CHFN_PROGRAM
-#define CHFN_PROGRAM "chfn"
-#endif
-#ifndef CHSH_PROGRAM
-#define CHSH_PROGRAM "chsh"
-#endif
 /*
  * exit status values
  */
@@ -105,7 +101,20 @@ static long inact = 0;		/* Days without change before locked */
 static int do_update_age = 0;
 
 #ifndef USE_PAM
-static char crypt_passwd[128];	/* The "old-style" password, if present */
+/*
+ * Size of the biggest passwd:
+ *   $6$	3
+ *   rounds=	7
+ *   999999999	9
+ *   $		1
+ *   salt	16
+ *   $		1
+ *   SHA512	123
+ *   nul	1
+ *
+ *   total	161
+ */
+static char crypt_passwd[256];
 static int do_update_pwd = 0;
 #endif
 
@@ -133,32 +142,38 @@ static void update_noshadow (void);
 
 static void update_shadow (void);
 static long getnumber (const char *);
+#ifdef WITH_SELINUX
+static int check_selinux_access (const char *changed_user,
+                                 uid_t changed_uid,
+                                 access_vector_t requested_access);
+#endif
 
 /*
  * usage - print command usage and exit
  */
 static void usage (int status)
 {
-	fprintf (stderr, _("Usage: passwd [options] [login]\n"
-			   "\n"
-			   "Options:\n"
-			   "  -a, --all			report password status on all accounts\n"
-			   "  -d, --delete			delete the password for the named account\n"
-			   "  -e, --expire			force expire the password for the named account\n"
-			   "  -h, --help			display this help message and exit\n"
-			   "  -k, --keep-tokens		change password only if expired\n"
-			   "  -i, --inactive INACTIVE	set password inactive after expiration\n"
-			   "				to INACTIVE\n"
-			   "  -l, --lock			lock the named account\n"
-			   "  -n, --mindays MIN_DAYS	set minimum number of days before password\n"
-			   "				change to MIN_DAYS\n"
-			   "  -q, --quiet			quiet mode\n"
-			   "  -r, --repository REPOSITORY	change password in REPOSITORY repository\n"
-			   "  -S, --status			report password status on the named account\n"
-			   "  -u, --unlock			unlock the named account\n"
-			   "  -w, --warndays WARN_DAYS	set expiration warning days to WARN_DAYS\n"
-			   "  -x, --maxdays MAX_DAYS	set maximim number of days before password\n"
-			   "				change to MAX_DAYS\n"));
+	fputs (_("Usage: passwd [options] [LOGIN]\n"
+	         "\n"
+	         "Options:\n"
+	         "  -a, --all                     report password status on all accounts\n"
+	         "  -d, --delete                  delete the password for the named account\n"
+	         "  -e, --expire                  force expire the password for the named account\n"
+	         "  -h, --help                    display this help message and exit\n"
+	         "  -k, --keep-tokens             change password only if expired\n"
+	         "  -i, --inactive INACTIVE       set password inactive after expiration\n"
+	         "                                to INACTIVE\n"
+	         "  -l, --lock                    lock the named account\n"
+	         "  -n, --mindays MIN_DAYS        set minimum number of days before password\n"
+	         "                                change to MIN_DAYS\n"
+	         "  -q, --quiet                   quiet mode\n"
+	         "  -r, --repository REPOSITORY   change password in REPOSITORY repository\n"
+	         "  -S, --status                  report password status on the named account\n"
+	         "  -u, --unlock                  unlock the named account\n"
+	         "  -w, --warndays WARN_DAYS      set expiration warning days to WARN_DAYS\n"
+	         "  -x, --maxdays MAX_DAYS        set maximim number of days before password\n"
+	         "                                change to MAX_DAYS\n"
+	         "\n"), stderr);
 	exit (status);
 }
 
@@ -198,7 +213,8 @@ static int new_password (const struct passwd *pw)
 	char pass[200];		/* New password */
 	int i;			/* Counter for retries */
 	int warned;
-	int pass_max_len;
+	int pass_max_len = -1;
+	char *method;
 
 #ifdef HAVE_LIBCRACK_HIST
 	int HistUpdate (const char *, const char *);
@@ -236,15 +252,33 @@ static int new_password (const struct passwd *pw)
 	 * for strength, unless it is the root user. This provides an escape
 	 * for initial login passwords.
 	 */
-	if (getdef_bool ("MD5_CRYPT_ENAB"))
-		pass_max_len = 127;
-	else
-		pass_max_len = getdef_num ("PASS_MAX_LEN", 8);
-
-	if (!qflg)
-		printf (_("\
-Enter the new password (minimum of %d, maximum of %d characters)\n\
-Please use a combination of upper and lower case letters and numbers.\n"), getdef_num ("PASS_MIN_LEN", 5), pass_max_len);
+	if ((method = getdef_str ("ENCRYPT_METHOD")) == NULL) {
+		if (!getdef_bool ("MD5_CRYPT_ENAB"))
+			pass_max_len = getdef_num ("PASS_MAX_LEN", 8);
+	} else {
+		if (   !strcmp (method, "MD5")
+#ifdef USE_SHA_CRYPT
+		    || !strcmp (method, "SHA256")
+		    || !strcmp (method, "SHA512")
+#endif
+		    )
+			pass_max_len = -1;
+		else
+			pass_max_len = getdef_num ("PASS_MAX_LEN", 8);
+	}
+	if (!qflg) {
+		if (pass_max_len == -1) {
+			printf (_(
+"Enter the new password (minimum of %d characters)\n"
+"Please use a combination of upper and lower case letters and numbers.\n"),
+				getdef_num ("PASS_MIN_LEN", 5));
+		} else {
+			printf (_(
+"Enter the new password (minimum of %d, maximum of %d characters)\n"
+"Please use a combination of upper and lower case letters and numbers.\n"),
+				getdef_num ("PASS_MIN_LEN", 5), pass_max_len);
+		}
+	}
 
 	warned = 0;
 	for (i = getdef_num ("PASS_CHANGE_TRIES", 5); i > 0; i--) {
@@ -258,7 +292,7 @@ Please use a combination of upper and lower case letters and numbers.\n"), getde
 		strzero (cp);
 
 		if (!amroot && (!obscure (orig, pass, pw) || reuse (pass, pw))) {
-			printf (_("Try again.\n"));
+			puts (_("Try again."));
 			continue;
 		}
 
@@ -269,8 +303,7 @@ Please use a combination of upper and lower case letters and numbers.\n"), getde
 		 */
 		if (amroot && !warned && getdef_bool ("PASS_ALWAYS_WARN")
 		    && (!obscure (orig, pass, pw) || reuse (pass, pw))) {
-			printf (_
-				("\nWarning: weak password (enter it again to use it anyway).\n"));
+			puts (_("\nWarning: weak password (enter it again to use it anyway)."));
 			warned++;
 			continue;
 		}
@@ -279,7 +312,7 @@ Please use a combination of upper and lower case letters and numbers.\n"), getde
 			return -1;
 		}
 		if (strcmp (cp, pass))
-			fprintf (stderr, _("They don't match; try again.\n"));
+			fputs (_("They don't match; try again.\n"), stderr);
 		else {
 			strzero (cp);
 			break;
@@ -295,7 +328,7 @@ Please use a combination of upper and lower case letters and numbers.\n"), getde
 	/*
 	 * Encrypt the password, then wipe the cleartext password.
 	 */
-	cp = pw_encrypt (pass, crypt_make_salt ());
+	cp = pw_encrypt (pass, crypt_make_salt (NULL, NULL));
 	memzero (pass, sizeof pass);
 
 #ifdef HAVE_LIBCRACK_HIST
@@ -358,7 +391,7 @@ static void check_password (const struct passwd *pw, const struct spwd *sp)
 	if (now < ok) {
 		fprintf (stderr,
 			 _
-			 ("Sorry, the password for %s cannot be changed yet.\n"),
+			 ("The password for %s cannot be changed yet.\n"),
 			 pw->pw_name);
 		SYSLOG ((LOG_WARN, "now < minimum age for `%s'", pw->pw_name));
 		closelog ();
@@ -407,7 +440,7 @@ static void print_status (const struct passwd *pw)
 {
 	struct spwd *sp;
 
-	sp = getspnam (pw->pw_name);
+	sp = getspnam (pw->pw_name); /* local, no need for xgetspnam */
 	if (sp) {
 		printf ("%s %s %s %ld %ld %ld %ld\n",
 			pw->pw_name,
@@ -444,10 +477,18 @@ static char *update_crypt_pw (char *cp)
 #endif
 
 	if (dflg)
-		cp = "";	/* XXX warning: const */
+		*cp = '\0';
 
-	if (uflg && *cp == '!')
-		cp++;
+	if (uflg && *cp == '!') {
+		if (cp[1] == '\0') {
+			fprintf (stderr,
+				 _("%s: unlocking the user would result in a passwordless account.\n"
+				   "You should set a password with usermod -p to unlock this user account.\n"),
+				 Prog);
+		} else {
+			cp++;
+		}
+        }
 
 	if (lflg && *cp != '!') {
 		char *newpw = xmalloc (strlen (cp) + 2);
@@ -466,14 +507,13 @@ static void update_noshadow (void)
 	struct passwd *npw;
 
 	if (!pw_lock ()) {
-		fprintf (stderr,
-			 _
-			 ("Cannot lock the password file; try again later.\n"));
+		fputs (_("Cannot lock the password file; try again later.\n"),
+		       stderr);
 		SYSLOG ((LOG_WARN, "can't lock password file"));
 		exit (E_PWDBUSY);
 	}
 	if (!pw_open (O_RDWR)) {
-		fprintf (stderr, _("Cannot open the password file.\n"));
+		fputs (_("Cannot open the password file.\n"), stderr);
 		SYSLOG ((LOG_ERR, "can't open password file"));
 		fail_exit (E_MISSING);
 	}
@@ -488,12 +528,12 @@ static void update_noshadow (void)
 		oom ();
 	npw->pw_passwd = update_crypt_pw (npw->pw_passwd);
 	if (!pw_update (npw)) {
-		fprintf (stderr, _("Error updating the password entry.\n"));
+		fputs (_("Error updating the password entry.\n"), stderr);
 		SYSLOG ((LOG_ERR, "error updating password entry"));
 		fail_exit (E_FAILURE);
 	}
 	if (!pw_close ()) {
-		fprintf (stderr, _("Cannot commit password file changes.\n"));
+		fputs (_("Cannot commit password file changes.\n"), stderr);
 		SYSLOG ((LOG_ERR, "can't rewrite password file"));
 		fail_exit (E_FAILURE);
 	}
@@ -506,14 +546,13 @@ static void update_shadow (void)
 	struct spwd *nsp;
 
 	if (!spw_lock ()) {
-		fprintf (stderr,
-			 _
-			 ("Cannot lock the password file; try again later.\n"));
+		fputs (_("Cannot lock the password file; try again later.\n"),
+		       stderr);
 		SYSLOG ((LOG_WARN, "can't lock password file"));
 		exit (E_PWDBUSY);
 	}
 	if (!spw_open (O_RDWR)) {
-		fprintf (stderr, _("Cannot open the password file.\n"));
+		fputs (_("Cannot open the password file.\n"), stderr);
 		SYSLOG ((LOG_ERR, "can't open password file"));
 		fail_exit (E_FAILURE);
 	}
@@ -539,6 +578,15 @@ static void update_shadow (void)
 		nsp->sp_inact = (inact * DAY) / SCALE;
 	if (do_update_age)
 		nsp->sp_lstchg = time ((time_t *) 0) / SCALE;
+	if (lflg) {
+		/* Set the account expiry field to 1.
+		 * Some PAM implementation consider zero as a non expired
+		 * account.
+		 */
+		nsp->sp_expire = 1;
+	}
+	if (uflg)
+		nsp->sp_expire = -1;
 
 	/*
 	 * Force change on next login, like SunOS 4.x passwd -e or Solaris
@@ -549,29 +597,79 @@ static void update_shadow (void)
 		nsp->sp_lstchg = 0;
 
 	if (!spw_update (nsp)) {
-		fprintf (stderr, _("Error updating the password entry.\n"));
+		fputs (_("Error updating the password entry.\n"), stderr);
 		SYSLOG ((LOG_ERR, "error updating password entry"));
 		fail_exit (E_FAILURE);
 	}
 	if (!spw_close ()) {
-		fprintf (stderr, _("Cannot commit password file changes.\n"));
+		fputs (_("Cannot commit password file changes.\n"), stderr);
 		SYSLOG ((LOG_ERR, "can't rewrite password file"));
 		fail_exit (E_FAILURE);
 	}
 	spw_unlock ();
 }
 
-static long getnumber (const char *str)
+static long getnumber (const char *numstr)
 {
 	long val;
-	char *cp;
+	char *errptr;
 
-	val = strtol (str, &cp, 10);
-	if (*cp)
-		usage (E_BAD_ARG);
+	val = strtol (numstr, &errptr, 10);
+	if (*errptr || errno == ERANGE) {
+		fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog,
+			 numstr);
+		exit (E_BAD_ARG);
+	}
 	return val;
 }
 
+#ifdef WITH_SELINUX
+static int check_selinux_access (const char *changed_user,
+                                 uid_t changed_uid,
+                                 access_vector_t requested_access)
+{
+	int status = -1;
+	security_context_t user_context;
+	const char *user;
+
+	/* if in permissive mode then allow the operation */
+	if (security_getenforce() == 0)
+		return 0;
+
+	/* get the context of the process which executed passwd */
+	if (getprevcon(&user_context))
+		return -1;
+
+	/* get the "user" portion of the context (the part before the first
+	   colon) */
+	context_t c;
+	c = context_new(user_context);
+	user = context_user_get(c);
+
+	/* if changing a password for an account with UID==0 or for an account
+	   where the identity matches then return success */
+	if (changed_uid != 0 && strcmp(changed_user, user) == 0) {
+		status = 0;
+	} else {
+		struct av_decision avd;
+		int retval;
+		retval = security_compute_av(user_context,
+		                             user_context,
+		                             SECCLASS_PASSWD,
+		                             requested_access,
+		                             &avd);
+		if ((retval == 0) &&
+		    ((requested_access & avd.allowed) == requested_access)) {
+			status = 0;
+		}
+	}
+	context_free(c);
+	freecon(user_context);
+	return status;
+}
+
+#endif
+
 /*
  * passwd - change a user's password file information
  *
@@ -650,7 +748,7 @@ int main (int argc, char **argv)
 			{"repository", required_argument, NULL, 'r'},
 			{"status", no_argument, NULL, 'S'},
 			{"unlock", no_argument, NULL, 'u'},
-			{"warning", required_argument, NULL, 'w'},
+			{"warndays", required_argument, NULL, 'w'},
 			{"maxdays", required_argument, NULL, 'x'},
 			{NULL, 0, NULL, '\0'}
 		};
@@ -744,6 +842,12 @@ int main (int argc, char **argv)
 	else
 		name = myname;
 
+	/*
+	 * Make sure that at most one username was specified.
+	 */
+	if (argc > optind+1)
+		usage (E_USAGE);
+
 	/*
 	 * The -a flag requires -S, no other flags, no username, and
 	 * you must be root.  --marekm
@@ -756,8 +860,10 @@ int main (int argc, char **argv)
 			exit (E_NOPERM);
 		}
 		setpwent ();
-		while ((pw = getpwent ()))
+		while ( (pw = getpwent ()) != NULL ) {
 			print_status (pw);
+		}
+		endpwent ();
 		exit (E_SUCCESS);
 	}
 #if 0
@@ -791,27 +897,36 @@ int main (int argc, char **argv)
 		exit (E_NOPERM);
 	}
 
-	pw = getpwnam (name);
+	pw = xgetpwnam (name);
 	if (!pw) {
 		fprintf (stderr, _("%s: unknown user %s\n"), Prog, name);
 		exit (E_NOPERM);
 	}
 #ifdef WITH_SELINUX
-	/*
-	 * If the UID of the user does not match the current real UID,
-	 * check if the change is allowed by SELinux policy.
-	 */
-	if ((pw->pw_uid != getuid ())
-	    && (is_selinux_enabled () > 0 ?
-		(selinux_check_passwd_access (PASSWD__PASSWD) != 0) :
-		!amroot)) {
-#else
+	/* only do this check when getuid()==0 because it's a pre-condition for
+	   changing a password without entering the old one */
+	if ((is_selinux_enabled() > 0) && (getuid() == 0) &&
+	    (check_selinux_access (name, pw->pw_uid, PASSWD__PASSWD) != 0)) {
+		security_context_t user_context;
+		if (getprevcon(&user_context) < 0) {
+			user_context = strdup("Unknown user context");
+		}
+		syslog(LOG_ALERT,
+		       "%s is not authorized to change the password of %s",
+		       user_context, name);
+		fprintf(stderr,
+		        _("%s: %s is not authorized to change the password of %s\n"),
+		        Prog, user_context, name);
+		freecon(user_context);
+		exit(1);
+	}
+#endif /* WITH_SELINUX */
+
 	/*
 	 * If the UID of the user does not match the current real UID,
 	 * check if I'm root.
 	 */
 	if (!amroot && pw->pw_uid != getuid ()) {
-#endif
 		fprintf (stderr,
 			 _
 			 ("%s: You may not view or modify password information for %s.\n"),
@@ -831,7 +946,7 @@ int main (int argc, char **argv)
 	/*
 	 * The user name is valid, so let's get the shadow file entry.
 	 */
-	sp = getspnam (name);
+	sp = getspnam (name); /* !USE_PAM, no need for xgetspnam */
 	if (!sp)
 		sp = pwd_to_spwd (pw);
 
@@ -884,7 +999,7 @@ int main (int argc, char **argv)
 	}
 #endif				/* USE_PAM */
 	if (setuid (0)) {
-		fprintf (stderr, _("Cannot change ID to root.\n"));
+		fputs (_("Cannot change ID to root.\n"), stderr);
 		SYSLOG ((LOG_ERR, "can't setuid(0)"));
 		closelog ();
 		exit (E_NOPERM);
@@ -901,9 +1016,9 @@ int main (int argc, char **argv)
 	closelog ();
 	if (!qflg) {
 		if (!eflg)
-			printf (_("Password changed.\n"));
+			puts (_("Password changed."));
 		else
-			printf (_("Password set to expire.\n"));
+			puts (_("Password set to expire."));
 	}
 	exit (E_SUCCESS);
 	/* NOT REACHED */
