@@ -1,5 +1,8 @@
 /*
- * Copyright 1991 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1991 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2000 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,27 +13,28 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: useradd.c,v 1.99 2006/06/22 11:30:32 kloczek Exp $"
+#ident "$Id: useradd.c 1986 2008-05-19 20:31:48Z nekral-guest $"
 
+#include <assert.h>
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -106,7 +110,11 @@ static int is_shadow_pwd;
 
 #ifdef SHADOWGRP
 static int is_shadow_grp;
+static int gshadow_locked = 0;
 #endif
+static int passwd_locked = 0;
+static int group_locked = 0;
+static int shadow_locked = 0;
 static char **user_groups;	/* NULL-terminated list */
 static long sys_ngroups;
 static int do_grp_update = 0;	/* group files need to be updated */
@@ -114,7 +122,7 @@ static int do_grp_update = 0;	/* group files need to be updated */
 static char *Prog;
 
 static int
- bflg = 0,			/* new default root of home directory */
+    bflg = 0,			/* new default root of home directory */
     cflg = 0,			/* comment (GECOS) field for new account */
     dflg = 0,			/* home directory for new account */
     Dflg = 0,			/* set/show new user default values */
@@ -123,14 +131,14 @@ static int
     gflg = 0,			/* primary group ID for new account */
     Gflg = 0,			/* secondary group set for new account */
     kflg = 0,			/* specify a directory to fill new user directory */
+    lflg = 0,			/* do not add user to lastlog database file */
     mflg = 0,			/* create user's home directory if it doesn't exist */
-    nflg = 0,			/* create a group having the same name as the user */
+    Nflg = 0,			/* do not create a group having the same name as the user, but add the user to def_group (or the group specified with -g) */
     oflg = 0,			/* permit non-unique user ID to be specified with -u */
+    rflg = 0,			/* create a system account */
     sflg = 0,			/* shell program for new account */
-    uflg = 0;			/* specify user ID for new account */
-
-extern char *optarg;
-extern int optind;
+    uflg = 0,			/* specify user ID for new account */
+    Uflg = 0;			/* create a group having the same name as the user */
 
 static int home_added;
 
@@ -171,7 +179,6 @@ static void new_pwent (struct passwd *);
 static long scale_age (long);
 static void new_spent (struct spwd *);
 static void grp_update (void);
-static void find_new_uid (void);
 
 static void process_flags (int argc, char **argv);
 static void close_files (void);
@@ -190,6 +197,21 @@ static void fail_exit (int code)
 	if (home_added)
 		rmdir (user_home);
 
+	if (shadow_locked) {
+		spw_unlock ();
+	}
+	if (passwd_locked) {
+		pw_unlock ();
+	}
+	if (group_locked) {
+		gr_unlock ();
+	}
+#ifdef	SHADOWGRP
+	if (gshadow_locked) {
+		sgr_unlock ();
+	}
+#endif
+
 #ifdef WITH_AUDIT
 	audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "adding user", user_name, -1,
 		      0);
@@ -198,42 +220,44 @@ static void fail_exit (int code)
 	exit (code);
 }
 
-static struct group *getgr_nam_gid (const char *name)
+static struct group *getgr_nam_gid (const char *grname)
 {
-	gid_t gid;
-	char *ep;
+	long gid;
+	char *errptr;
 
-	gid = strtoul (name, &ep, 10);
-	if (*name != '\0' && *ep == '\0')	/* valid numeric GID */
-		return getgrgid (gid);
-
-	return getgrnam (name);
+	gid = strtol (grname, &errptr, 10);
+	if (*grname != '\0' && *errptr == '\0' && errno != ERANGE && gid >= 0)
+		return xgetgrgid (gid);
+	return xgetgrnam (grname);
 }
 
-static long get_number (const char *cp)
+static long get_number (const char *numstr)
 {
 	long val;
-	char *ep;
+	char *errptr;
 
-	val = strtol (cp, &ep, 10);
-	if (*cp != '\0' && *ep == '\0')	/* valid number */
-		return val;
-
-	fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog, cp);
-	exit (E_BAD_ARG);
+	val = strtol (numstr, &errptr, 10);
+	if (*errptr || errno == ERANGE) {
+		fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog,
+			 numstr);
+		exit (E_BAD_ARG);
+	}
+	return val;
 }
 
-static uid_t get_uid (const char *cp)
+static uid_t get_uid (const char *uidstr)
 {
-	uid_t val;
-	char *ep;
+	long val;
+	char *errptr;
 
-	val = strtoul (cp, &ep, 10);
-	if (*cp != '\0' && *ep == '\0')	/* valid number */
-		return val;
-
-	fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog, cp);
-	exit (E_BAD_ARG);
+	val = strtol (uidstr, &errptr, 10);
+	if (*errptr || errno == ERANGE || val < 0) {
+		fprintf (stderr,
+			 _("%s: invalid numeric argument '%s'\n"), Prog,
+			 uidstr);
+		exit (E_BAD_ARG);
+	}
+	return val;
 }
 
 #define MATCH(x,y) (strncmp((x),(y),strlen(y)) == 0)
@@ -250,7 +274,6 @@ static void get_defaults (void)
 	FILE *fp;
 	char buf[1024];
 	char *cp, *ep;
-	const struct group *grp;
 
 	/*
 	 * Open the defaults file for reading.
@@ -277,9 +300,11 @@ static void get_defaults (void)
 		 */
 		if (MATCH (buf, DGROUP)) {
 			unsigned int val = (unsigned int) strtoul (cp, &ep, 10);
+			const struct group *grp;
 
-			if (*cp != '\0' && *ep == '\0') {	/* valid number */
+			if (*cp != '\0' && *ep == '\0') { /* valid number */
 				def_group = val;
+				/* local, no need for xgetgrgid */
 				if ((grp = getgrgid (def_group))) {
 					def_gname = xstrdup (grp->gr_name);
 				} else {
@@ -287,6 +312,7 @@ static void get_defaults (void)
 						 _("%s: unknown GID %s\n"),
 						 Prog, cp);
 				}
+			/* local, no need for xgetgrnam */
 			} else if ((grp = getgrnam (cp))) {
 				def_group = grp->gr_gid;
 				def_gname = xstrdup (cp);
@@ -316,7 +342,7 @@ static void get_defaults (void)
 		else if (MATCH (buf, INACT)) {
 			long val = strtol (cp, &ep, 10);
 
-			if (*cp != '\0' && *ep == '\0')	/* valid number */
+			if (*cp || errno == ERANGE)
 				def_inactive = val;
 			else
 				def_inactive = -1;
@@ -608,33 +634,39 @@ static int get_groups (char *list)
  */
 static void usage (void)
 {
-	fprintf (stderr, _("Usage: useradd [options] LOGIN\n"
-			   "\n"
-			   "Options:\n"
-			   "  -b, --base-dir BASE_DIR	base directory for the new user account\n"
-			   "				home directory\n"
-			   "  -c, --comment COMMENT		set the GECOS field for the new user account\n"
-			   "  -d, --home-dir HOME_DIR	home directory for the new user account\n"
-			   "  -D, --defaults		print or save modified default useradd\n"
-			   "				configuration\n"
-			   "  -e, --expiredate EXPIRE_DATE	set account expiration date to EXPIRE_DATE\n"
-			   "  -f, --inactive INACTIVE	set password inactive after expiration\n"
-			   "				to INACTIVE\n"
-			   "  -g, --gid GROUP		force use GROUP for the new user account\n"
-			   "  -G, --groups GROUPS		list of supplementary groups for the new\n"
-			   "				user account\n"
-			   "  -h, --help			display this help message and exit\n"
-			   "  -k, --skel SKEL_DIR		specify an alternative skel directory\n"
-			   "  -K, --key KEY=VALUE		overrides /etc/login.defs defaults\n"
-			   "  -m, --create-home		create home directory for the new user\n"
-			   "				account\n"
-			   "  -o, --non-unique		allow create user with duplicate\n"
-			   "				(non-unique) UID\n"
-			   "  -p, --password PASSWORD	use encrypted password for the new user\n"
-			   "				account\n"
-			   "  -s, --shell SHELL		the login shell for the new user account\n"
-			   "  -u, --uid UID			force use the UID for the new user account\n"
-			   "\n"));
+	fputs (_("Usage: useradd [options] LOGIN\n"
+	         "\n"
+	         "Options:\n"
+	         "  -b, --base-dir BASE_DIR       base directory for the new user account\n"
+	         "                                home directory\n"
+	         "  -c, --comment COMMENT         set the GECOS field for the new user account\n"
+	         "  -d, --home-dir HOME_DIR       home directory for the new user account\n"
+	         "  -D, --defaults                print or save modified default useradd\n"
+	         "                                configuration\n"
+	         "  -e, --expiredate EXPIRE_DATE  set account expiration date to EXPIRE_DATE\n"
+	         "  -f, --inactive INACTIVE       set password inactive after expiration\n"
+	         "                                to INACTIVE\n"
+	         "  -g, --gid GROUP               force use GROUP for the new user account\n"
+	         "  -G, --groups GROUPS           list of supplementary groups for the new\n"
+	         "                                user account\n"
+	         "  -h, --help                    display this help message and exit\n"
+	         "  -k, --skel SKEL_DIR           specify an alternative skel directory\n"
+	         "  -K, --key KEY=VALUE           overrides /etc/login.defs defaults\n"
+	         "  -l,                           do not add the user to the lastlog and\n"
+	         "                                faillog databases\n"
+	         "  -m, --create-home             create home directory for the new user\n"
+	         "                                account\n"
+	         "  -N, --no-user-group           do not create a group with the same name as\n"
+	         "                                the user\n"
+	         "  -o, --non-unique              allow create user with duplicate\n"
+	         "                                (non-unique) UID\n"
+	         "  -p, --password PASSWORD       use encrypted password for the new user\n"
+	         "                                account\n"
+	         "  -r, --system                  create a system account\n"
+	         "  -s, --shell SHELL             the login shell for the new user account\n"
+	         "  -u, --uid UID                 force use the UID for the new user account\n"
+	         "  -U, --user-group              create a group with the same name as the user\n"
+	         "\n"), stderr);
 	exit (E_USAGE);
 }
 
@@ -680,11 +712,19 @@ static void new_spent (struct spwd *spent)
 	spent->sp_namp = (char *) user_name;
 	spent->sp_pwdp = (char *) user_pass;
 	spent->sp_lstchg = time ((time_t *) 0) / SCALE;
-	spent->sp_min = scale_age (getdef_num ("PASS_MIN_DAYS", -1));
-	spent->sp_max = scale_age (getdef_num ("PASS_MAX_DAYS", -1));
-	spent->sp_warn = scale_age (getdef_num ("PASS_WARN_AGE", -1));
-	spent->sp_inact = scale_age (def_inactive);
-	spent->sp_expire = scale_age (user_expire);
+	if (!rflg) {
+		spent->sp_min = scale_age (getdef_num ("PASS_MIN_DAYS", -1));
+		spent->sp_max = scale_age (getdef_num ("PASS_MAX_DAYS", -1));
+		spent->sp_warn = scale_age (getdef_num ("PASS_WARN_AGE", -1));
+		spent->sp_inact = scale_age (def_inactive);
+		spent->sp_expire = scale_age (user_expire);
+	} else {
+		spent->sp_min = scale_age (-1);
+		spent->sp_max = scale_age (-1);
+		spent->sp_warn = scale_age (-1);
+		spent->sp_inact = scale_age (-1);
+		spent->sp_expire = scale_age (-1);
+	}
 	spent->sp_flag = -1;
 }
 
@@ -693,6 +733,11 @@ static void new_spent (struct spwd *spent)
  *
  *	grp_update() takes the secondary group set given in user_groups
  *	and adds the user to each group given by that set.
+ *
+ *	The group files are opened and locked in open_files().
+ *
+ *	close_files() should be called afterwards to commit the changes
+ *	and unlocking the group files.
  */
 static void grp_update (void)
 {
@@ -704,32 +749,6 @@ static void grp_update (void)
 	struct sgrp *nsgrp;
 #endif
 
-	/*
-	 * Test for unique entries of user_groups in /etc/group
-	 * pvrabec@redhat.com
-	 */
-	char **user_groups_tmp = user_groups;
-
-	while (*user_groups_tmp) {
-		int count = 0;
-
-		for (gr_rewind (), grp = gr_next (); grp && count < 2;
-		     grp = gr_next ()) {
-			if (strcmp (*user_groups_tmp, grp->gr_name) == 0) {
-				count++;
-			}
-		}
-		if (count > 1) {
-			fprintf (stderr,
-				 "%s: error not unique group names in group file\n",
-				 Prog);
-			fail_exit (E_GRP_UPDATE);
-		}
-		user_groups_tmp++;
-	}
-
-	/* Locking and opening of the group files moved to open_files() --gafton */
-
 	/*
 	 * Scan through the entire group file looking for the groups that
 	 * the user is a member of.
@@ -749,6 +768,9 @@ static void grp_update (void)
 		 */
 		ngrp = __gr_dup (grp);
 		if (!ngrp) {
+			fprintf (stderr,
+				 _("%s: Out of memory. Cannot update the group database.\n"),
+				 Prog);
 			fail_exit (E_GRP_UPDATE);	/* XXX */
 		}
 
@@ -797,6 +819,9 @@ static void grp_update (void)
 		 */
 		nsgrp = __sgr_dup (sgrp);
 		if (!nsgrp) {
+			fprintf (stderr,
+				 _("%s: Out of memory. Cannot update the shadow group database.\n"),
+				 Prog);
 			fail_exit (E_GRP_UPDATE);	/* XXX */
 		}
 
@@ -820,165 +845,6 @@ static void grp_update (void)
 #endif				/* SHADOWGRP */
 }
 
-/*
- * find_new_uid - find the next available UID
- *
- *	find_new_uid() locates the next highest unused UID in the password
- *	file, or checks the given user ID against the existing ones for
- *	uniqueness.
- */
-static void find_new_uid (void)
-{
-	const struct passwd *pwd;
-	uid_t uid_min, uid_max;
-
-	uid_min = getdef_unum ("UID_MIN", 1000);
-	uid_max = getdef_unum ("UID_MAX", 60000);
-
-	/*
-	 * Start with some UID value if the user didn't provide us with
-	 * one already.
-	 */
-	if (!uflg)
-		user_id = uid_min;
-
-	/*
-	 * Search the entire password file, either looking for this
-	 * UID (if the user specified one with -u) or looking for the
-	 * largest unused value.
-	 */
-#ifdef NO_GETPWENT
-	pw_rewind ();
-	while ((pwd = pw_next ())) {
-#else				/* using getpwent() we can check against NIS users etc. */
-	setpwent ();
-	while ((pwd = getpwent ())) {
-#endif
-		if (strcmp (user_name, pwd->pw_name) == 0) {
-			fprintf (stderr, _("%s: name %s is not unique\n"),
-				 Prog, user_name);
-#ifdef WITH_AUDIT
-			audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "adding user",
-				      user_name, user_id, 0);
-#endif
-			exit (E_NAME_IN_USE);
-		}
-		if (uflg && user_id == pwd->pw_uid) {
-			fprintf (stderr, _("%s: UID %u is not unique\n"),
-				 Prog, (unsigned int) user_id);
-#ifdef WITH_AUDIT
-			audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "adding user",
-				      user_name, user_id, 0);
-#endif
-			exit (E_UID_IN_USE);
-		}
-		if (!uflg && pwd->pw_uid >= user_id) {
-			if (pwd->pw_uid > uid_max)
-				continue;
-			user_id = pwd->pw_uid + 1;
-		}
-	}
-
-	/*
-	 * If a user with UID equal to UID_MAX exists, the above algorithm
-	 * will give us UID_MAX+1 even if not unique. Search for the first
-	 * free UID starting with UID_MIN (it's O(n*n) but can be avoided
-	 * by not having users with UID equal to UID_MAX).  --marekm
-	 */
-	if (!uflg && user_id == uid_max + 1) {
-		for (user_id = uid_min; user_id < uid_max; user_id++) {
-#ifdef NO_GETPWENT
-			pw_rewind ();
-			while ((pwd = pw_next ())
-			       && pwd->pw_uid != user_id);
-			if (!pwd)
-				break;
-#else
-			if (!getpwuid (user_id))
-				break;
-#endif
-		}
-		if (user_id == uid_max) {
-			fprintf (stderr, _("%s: can't get unique UID\n"), Prog);
-			fail_exit (E_UID_IN_USE);
-		}
-	}
-}
-
- /*
-  * find_new_gid - find the next available GID
-  *
-  *     find_new_gid() locates the next highest unused GID in the group
-  *     file, or checks the given group ID against the existing ones for
-  *     uniqueness.
-  */
-
-static void find_new_gid ()
-{
-	const struct group *grp;
-	gid_t gid_min, gid_max;
-
-	gid_min = getdef_num ("GID_MIN", 500);
-	gid_max = getdef_num ("GID_MAX", 60000);
-
-	/*
-	 * Start with some GID value if the user didn't provide us with
-	 * one already.
-	 */
-	user_gid = gid_min;
-
-	/*
-	 * Search the entire group file, either looking for this
-	 * GID (if the user specified one with -g) or looking for the
-	 * largest unused value.
-	 */
-#ifdef NO_GETGRENT
-	gr_rewind ();
-	while ((grp = gr_next ()))
-#else
-	setgrent ();
-	while ((grp = getgrent ()))
-#endif
-	{
-		if (strcmp (user_name, grp->gr_name) == 0) {
-			user_gid = grp->gr_gid;
-			return;
-		}
-		if (grp->gr_gid >= user_gid) {
-			if (grp->gr_gid > gid_max)
-				continue;
-			user_gid = grp->gr_gid + 1;
-		}
-	}
-#ifndef NO_GETGRENT		/* glibc does have this, so ... */
-	/* A quick test gets here: if the UID is available
-	 * as a GID, go ahead and use it */
-	if (!getgrgid (user_id)) {
-		user_gid = user_id;
-		return;
-	}
-#endif
-	if (user_gid == gid_max + 1) {
-		for (user_gid = gid_min; user_gid < gid_max; user_gid++) {
-#ifdef NO_GETGRENT
-			gr_rewind ();
-			while ((grp = gr_next ()) && grp->gr_gid != user_gid);
-			if (!grp)
-				break;
-#else
-			if (!getgrgid (user_gid))
-				break;
-#endif
-		}
-		if (user_gid == gid_max) {
-			fprintf (stderr,
-				 "%s: can't get unique gid (run out of GIDs)\n",
-				 Prog);
-			fail_exit (4);
-		}
-	}
-}
-
 /*
  * process_flags - perform command line argument setting
  *
@@ -1001,7 +867,7 @@ static void process_flags (int argc, char **argv)
 			{"base-dir", required_argument, NULL, 'b'},
 			{"comment", required_argument, NULL, 'c'},
 			{"home-dir", required_argument, NULL, 'd'},
-			{"defaults", required_argument, NULL, 'D'},
+			{"defaults", no_argument, NULL, 'D'},
 			{"expiredate", required_argument, NULL, 'e'},
 			{"inactive", required_argument, NULL, 'f'},
 			{"gid", required_argument, NULL, 'g'},
@@ -1010,20 +876,20 @@ static void process_flags (int argc, char **argv)
 			{"skel", required_argument, NULL, 'k'},
 			{"key", required_argument, NULL, 'K'},
 			{"create-home", no_argument, NULL, 'm'},
+			{"no-user-group", no_argument, NULL, 'N'},
 			{"non-unique", no_argument, NULL, 'o'},
 			{"password", required_argument, NULL, 'p'},
+			{"system", no_argument, NULL, 'r'},
 			{"shell", required_argument, NULL, 's'},
 			{"uid", required_argument, NULL, 'u'},
+			{"user-group", no_argument, NULL, 'U'},
 			{NULL, 0, NULL, '\0'}
 		};
 		while ((c =
-			getopt_long (argc, argv, "b:c:d:De:f:g:G:k:K:mMop:s:u:",
+			getopt_long (argc, argv, "b:c:d:De:f:g:G:k:K:lmMNop:rs:u:U",
 				     long_options, NULL)) != -1) {
 			switch (c) {
 			case 'b':
-				if (!Dflg)
-					usage ();
-
 				if (!VALID (optarg)
 				    || optarg[0] != '/') {
 					fprintf (stderr,
@@ -1154,9 +1020,15 @@ static void process_flags (int argc, char **argv)
 				if (putdef_str (optarg, cp) < 0)
 					exit (E_BAD_ARG);
 				break;
+			case 'l':
+				lflg++;
+				break;
 			case 'm':
 				mflg++;
 				break;
+			case 'N':
+				Nflg++;
+				break;
 			case 'o':
 				oflg++;
 				break;
@@ -1170,6 +1042,9 @@ static void process_flags (int argc, char **argv)
 				}
 				user_pass = optarg;
 				break;
+			case 'r':
+				rflg++;
+				break;
 			case 's':
 				if (!VALID (optarg)
 				    || (optarg[0]
@@ -1189,6 +1064,9 @@ static void process_flags (int argc, char **argv)
 				user_id = get_uid (optarg);
 				uflg++;
 				break;
+			case 'U':
+				Uflg++;
+				break;
 			default:
 				usage ();
 			}
@@ -1196,12 +1074,39 @@ static void process_flags (int argc, char **argv)
 		}
 	}
 
+	if (!gflg && !Nflg && !Uflg) {
+		/* Get the settings from login.defs */
+		Uflg = getdef_bool ("USERGROUPS_ENAB");
+	}
+
 	/*
 	 * Certain options are only valid in combination with others.
 	 * Check it here so that they can be specified in any order.
 	 */
-	if ((oflg && !uflg) || (kflg && !mflg))
+	if (oflg && !uflg) {
+		fprintf (stderr,
+		         _("%s: %s flag is ONLY allowed with the %s flag\n"),
+		         Prog, "-o", "-u");
 		usage ();
+	}
+	if (kflg && !mflg) {
+		fprintf (stderr,
+		         _("%s: %s flag is ONLY allowed with the %s flag\n"),
+		         Prog, "-k", "-m");
+		usage ();
+	}
+	if (Uflg && gflg) {
+		fprintf (stderr,
+		         _("%s: options %s and %s conflict\n"),
+		         Prog, "-U", "-g");
+		usage ();
+	}
+	if (Uflg && Nflg) {
+		fprintf (stderr,
+		         _("%s: options %s and %s conflict\n"),
+		         Prog, "-U", "-N");
+		usage ();
+	}
 
 	/*
 	 * Either -D or username is required. Defaults can be set with -D
@@ -1287,13 +1192,19 @@ static void close_files (void)
 		}
 #endif
 	}
-	if (is_shadow_pwd)
+	if (is_shadow_pwd) {
 		spw_unlock ();
+		shadow_locked--;
+	}
 	pw_unlock ();
+	passwd_locked--;
 	gr_unlock ();
+	group_locked--;
 #ifdef	SHADOWGRP
-	if (is_shadow_grp)
+	if (is_shadow_grp) {
 		sgr_unlock ();
+		gshadow_locked--;
+	}
 #endif
 }
 
@@ -1312,60 +1223,68 @@ static void open_files (void)
 #endif
 		exit (E_PW_UPDATE);
 	}
+	passwd_locked++;
 	if (!pw_open (O_RDWR)) {
 		fprintf (stderr, _("%s: unable to open password file\n"), Prog);
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
 			      "opening password file", user_name, user_id, 0);
 #endif
-		pw_unlock ();
-		exit (E_PW_UPDATE);
+		fail_exit (E_PW_UPDATE);
 	}
-	if (is_shadow_pwd && !spw_lock ()) {
-		fprintf (stderr,
-			 _("%s: cannot lock shadow password file\n"), Prog);
+	if (is_shadow_pwd) {
+		if (!spw_lock ()) {
+			fprintf (stderr,
+			         _("%s: cannot lock shadow password file\n"),
+			         Prog);
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "locking shadow password file", user_name,
-			      user_id, 0);
+			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
+			              "locking shadow password file", user_name,
+			              user_id, 0);
 #endif
-		pw_unlock ();
-		exit (E_PW_UPDATE);
-	}
-	if (is_shadow_pwd && !spw_open (O_RDWR)) {
-		fprintf (stderr,
-			 _("%s: cannot open shadow password file\n"), Prog);
+			fail_exit (E_PW_UPDATE);
+		}
+		shadow_locked++;
+		if (!spw_open (O_RDWR)) {
+			fprintf (stderr,
+			         _("%s: cannot open shadow password file\n"),
+			         Prog);
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "opening shadow password file", user_name,
-			      user_id, 0);
+			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
+			              "opening shadow password file", user_name,
+			              user_id, 0);
 #endif
-		spw_unlock ();
-		pw_unlock ();
-		exit (E_PW_UPDATE);
+			fail_exit (E_PW_UPDATE);
+		}
 	}
+
 	/*
 	 * Lock and open the group file.
 	 */
-
 	if (!gr_lock ()) {
 		fprintf (stderr, _("%s: error locking group file\n"), Prog);
 		fail_exit (E_GRP_UPDATE);
 	}
+	group_locked++;
 	if (!gr_open (O_RDWR)) {
 		fprintf (stderr, _("%s: error opening group file\n"), Prog);
 		fail_exit (E_GRP_UPDATE);
 	}
 #ifdef  SHADOWGRP
-	if (is_shadow_grp && !sgr_lock ()) {
-		fprintf (stderr,
-			 _("%s: error locking shadow group file\n"), Prog);
-		fail_exit (E_GRP_UPDATE);
-	}
-	if (is_shadow_grp && !sgr_open (O_RDWR)) {
-		fprintf (stderr,
-			 _("%s: error opening shadow group file\n"), Prog);
-		fail_exit (E_GRP_UPDATE);
+	if (is_shadow_grp) {
+		if (!sgr_lock ()) {
+			fprintf (stderr,
+			         _("%s: error locking shadow group file\n"),
+			         Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+		gshadow_locked++;
+		if (!sgr_open (O_RDWR)) {
+			fprintf (stderr,
+			         _("%s: error opening shadow group file\n"),
+			         Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
 	}
 #endif
 }
@@ -1488,9 +1407,6 @@ static void usr_update (void)
 	struct passwd pwent;
 	struct spwd spent;
 
-	if (!oflg)
-		find_new_uid ();
-
 	/*
 	 * Fill in the password structure with any new fields, making
 	 * copies of strings.
@@ -1513,7 +1429,8 @@ static void usr_update (void)
 	 * no user with this UID exists yet (entries for shared UIDs
 	 * are left unchanged).  --marekm
 	 */
-	if (!getpwuid (user_id)) {
+	/* local, no need for xgetpwuid */
+	if ((!lflg) && (getpwuid (user_id) == NULL)) {
 		faillog_reset (user_id);
 		lastlog_reset (user_id);
 	}
@@ -1524,7 +1441,7 @@ static void usr_update (void)
 	if (!pw_update (&pwent)) {
 		fprintf (stderr,
 			 _("%s: error adding new password entry\n"), Prog);
-		exit (E_PW_UPDATE);
+		fail_exit (E_PW_UPDATE);
 	}
 
 	/*
@@ -1539,7 +1456,7 @@ static void usr_update (void)
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
 			      "adding shadow password", user_name, user_id, 0);
 #endif
-		exit (E_PW_UPDATE);
+		fail_exit (E_PW_UPDATE);
 	}
 #ifdef WITH_AUDIT
 	audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "adding user", user_name,
@@ -1569,6 +1486,11 @@ static void create_home (void)
 				 _
 				 ("%s: cannot create directory %s\n"),
 				 Prog, user_home);
+#ifdef WITH_AUDIT
+			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
+				      "adding home directory", user_name,
+				      user_id, 0);
+#endif
 			fail_exit (E_HOMEDIR);
 		}
 		chown (user_home, user_id, user_gid);
@@ -1598,7 +1520,10 @@ static void create_mail (void)
 	mode_t mode;
 
 	if (strcasecmp (create_mail_spool, "yes") == 0) {
-		spool = getdef_str ("MAIL_DIR") ? : "/var/mail";
+		spool = getdef_str ("MAIL_DIR");
+		if (NULL == spool) {
+			spool = "/var/mail";
+		}
 		file = alloca (strlen (spool) + strlen (user_name) + 2);
 		sprintf (file, "%s/%s", spool, user_name);
 		fd = open (file, O_CREAT | O_WRONLY | O_TRUNC | O_EXCL, 0);
@@ -1607,11 +1532,10 @@ static void create_mail (void)
 			return;
 		}
 
-		gr = getgrnam ("mail");
+		gr = getgrnam ("mail"); /* local, no need for xgetgrnam */
 		if (!gr) {
-			fprintf (stderr,
-				 _
-				 ("Group 'mail' not found. Creating the user mailbox file with 0600 mode.\n"));
+			fputs (_("Group 'mail' not found. Creating the user mailbox file with 0600 mode.\n"),
+			       stderr);
 			gid = user_gid;
 			mode = 0600;
 		} else {
@@ -1633,7 +1557,6 @@ int main (int argc, char **argv)
 {
 #ifdef USE_PAM
 	pam_handle_t *pamh = NULL;
-	struct passwd *pampw;
 	int retval;
 #endif
 
@@ -1672,13 +1595,17 @@ int main (int argc, char **argv)
 #ifdef USE_PAM
 	retval = PAM_SUCCESS;
 
-	pampw = getpwuid (getuid ());
-	if (pampw == NULL) {
-		retval = PAM_USER_UNKNOWN;
-	}
+	{
+		struct passwd *pampw;
+		pampw = getpwuid (getuid ()); /* local, no need for xgetpwuid */
+		if (pampw == NULL) {
+			retval = PAM_USER_UNKNOWN;
+		}
 
-	if (retval == PAM_SUCCESS) {
-		retval = pam_start ("useradd", pampw->pw_name, &conv, &pamh);
+		if (retval == PAM_SUCCESS) {
+			retval = pam_start ("useradd", pampw->pw_name,
+					    &conv, &pamh);
+		}
 	}
 
 	if (retval == PAM_SUCCESS) {
@@ -1697,7 +1624,7 @@ int main (int argc, char **argv)
 
 	if (retval != PAM_SUCCESS) {
 		fprintf (stderr, _("%s: PAM authentication failed\n"), Prog);
-		exit (1);
+		fail_exit (1);
 	}
 #endif				/* USE_PAM */
 
@@ -1716,13 +1643,13 @@ int main (int argc, char **argv)
 	/*
 	 * Start with a quick check to see if the user exists.
 	 */
-	if (getpwnam (user_name)) {
+	if (getpwnam (user_name)) { /* local, no need for xgetpwnam */
 		fprintf (stderr, _("%s: user %s exists\n"), Prog, user_name);
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "adding user",
 			      user_name, -1, 0);
 #endif
-		exit (E_NAME_IN_USE);
+		fail_exit (E_NAME_IN_USE);
 	}
 
 	/*
@@ -1731,8 +1658,8 @@ int main (int argc, char **argv)
 	 * to that group, use useradd -g username username.
 	 * --bero
 	 */
-	if (!gflg) {
-		if (getgrnam (user_name)) {
+	if (Uflg) {
+		if (getgrnam (user_name)) { /* local, no need for xgetgrnam */
 			fprintf (stderr,
 				 _
 				 ("%s: group %s exists - if you want to add this user to that group, use -g.\n"),
@@ -1741,7 +1668,7 @@ int main (int argc, char **argv)
 			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
 				      "adding group", user_name, -1, 0);
 #endif
-			exit (E_NAME_IN_USE);
+			fail_exit (E_NAME_IN_USE);
 		}
 	}
 
@@ -1760,13 +1687,31 @@ int main (int argc, char **argv)
 		/* first, seek for a valid uid to use for this user.
 		 * We do this because later we can use the uid we found as
 		 * gid too ... --gafton */
-		find_new_uid ();
+		if (!uflg) {
+			if (find_new_uid (rflg, &user_id, NULL) < 0) {
+				fprintf (stderr, _("%s: can't create user\n"), Prog);
+				fail_exit (E_UID_IN_USE);
+			}
+		} else {
+			if (getpwuid (user_id) != NULL) {
+				fprintf (stderr, _("%s: UID %u is not unique\n"), Prog, (unsigned int) user_id);
+#ifdef WITH_AUDIT
+				audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "adding user", user_name, user_id, 0);
+#endif
+				fail_exit (E_UID_IN_USE);
+			}
+		}
 	}
 
 	/* do we have to add a group for that user? This is why we need to
 	 * open the group files in the open_files() function  --gafton */
-	if (!(nflg || gflg)) {
-		find_new_gid ();
+	if (Uflg) {
+		if (find_new_gid (rflg, &user_gid, &user_id) < 0) {
+			fprintf (stderr,
+				 _("%s: can't create group\n"),
+				 Prog);
+			fail_exit (4);
+		}
 		grp_add ();
 	}
 
@@ -1804,16 +1749,15 @@ int main (int argc, char **argv)
 
 	create_mail ();
 
+	close_files ();
+
 	nscd_flush_cache ("passwd");
 	nscd_flush_cache ("group");
 
-	close_files ();
-
 #ifdef USE_PAM
 	if (retval == PAM_SUCCESS)
 		pam_end (pamh, PAM_SUCCESS);
 #endif				/* USE_PAM */
 
-	exit (E_SUCCESS);
-	/* NOT REACHED */
+	return E_SUCCESS;
 }
