@@ -1,5 +1,8 @@
 /*
- * Copyright 1991 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1991 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2000 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,29 +13,30 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: userdel.c,v 1.61 2006/02/07 20:19:46 kloczek Exp $"
+#ident "$Id: userdel.c 2007 2008-05-24 15:35:15Z nekral-guest $"
 
 #include <errno.h>
 #include <fcntl.h>
+#include <getopt.h>
 #include <grp.h>
 #include <pwd.h>
 #include <stdio.h>
@@ -98,7 +102,14 @@ static void remove_mailbox (void);
  */
 static void usage (void)
 {
-	fprintf (stderr, _("Usage: %s [-r] name\n"), Prog);
+	fputs (_("Usage: userdel [options] LOGIN\n"
+	         "\n"
+	         "Options:\n"
+	         "  -f, --force                   force removal of files,\n"
+	         "                                even if not owned by user\n"
+	         "  -h, --help                    display this help message and exit\n"
+	         "  -r, --remove                  remove home directory and mail spool\n"
+	         "\n"), stderr);
 	exit (E_USAGE);
 }
 
@@ -143,12 +154,17 @@ static void update_groups (void)
 		 */
 		ngrp = __gr_dup (grp);
 		if (!ngrp) {
+			fprintf (stderr,
+				 _("%s: Out of memory. Cannot update the group database.\n"),
+				 Prog);
 			exit (13);	/* XXX */
 		}
 		ngrp->gr_mem = del_list (ngrp->gr_mem, user_name);
-		if (!gr_update (ngrp))
+		if (!gr_update (ngrp)) {
 			fprintf (stderr,
 				 _("%s: error updating group entry\n"), Prog);
+			exit (E_GRP_UPDATE);
+		}
 
 		/*
 		 * Update the DBM group file with the new entry as well.
@@ -167,7 +183,7 @@ static void update_groups (void)
 	 * now if they have a group with the same name as their
 	 * user name, with no members, we delete it.
 	 */
-	grp = getgrnam (user_name);
+	grp = xgetgrnam (user_name);
 	if (grp && getdef_bool ("USERGROUPS_ENAB")
 	    && (grp->gr_mem[0] == NULL)) {
 
@@ -236,6 +252,9 @@ static void update_groups (void)
 
 		nsgrp = __sgr_dup (sgrp);
 		if (!nsgrp) {
+			fprintf (stderr,
+				 _("%s: Out of memory. Cannot update the shadow group database.\n"),
+				 Prog);
 			exit (13);	/* XXX */
 		}
 
@@ -245,9 +264,11 @@ static void update_groups (void)
 		if (was_admin)
 			nsgrp->sg_adm = del_list (nsgrp->sg_adm, user_name);
 
-		if (!sgr_update (nsgrp))
+		if (!sgr_update (nsgrp)) {
 			fprintf (stderr,
-				 _("%s: error updating group entry\n"), Prog);
+				 _("%s: error updating shadow group entry\n"), Prog);
+			exit (E_GRP_UPDATE);
+		}
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
 			      "deleting user from shadow group", user_name,
@@ -404,12 +425,16 @@ static void open_files (void)
  */
 static void update_user (void)
 {
-	if (!pw_remove (user_name))
+	if (!pw_remove (user_name)) {
 		fprintf (stderr,
 			 _("%s: error deleting password entry\n"), Prog);
-	if (is_shadow_pwd && !spw_remove (user_name))
+		fail_exit (E_PW_UPDATE);
+	}
+	if (is_shadow_pwd && !spw_remove (user_name)) {
 		fprintf (stderr,
 			 _("%s: error deleting shadow password entry\n"), Prog);
+		fail_exit (E_PW_UPDATE);
+	}
 #ifdef WITH_AUDIT
 	audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "deleting user entries",
 		      user_name, user_id, 1);
@@ -442,13 +467,9 @@ static void user_busy (const char *name, uid_t uid)
 	setutent ();
 	while ((utent = getutent ())) {
 #endif
-#ifdef USER_PROCESS
 		if (utent->ut_type != USER_PROCESS)
 			continue;
-#else
-		if (utent->ut_user[0] == '\0')
-			continue;
-#endif
+
 		if (strncmp (utent->ut_user, name, sizeof utent->ut_user))
 			continue;
 		fprintf (stderr,
@@ -525,7 +546,8 @@ static void user_cancel (const char *user)
 #ifdef EXTRA_CHECK_HOME_DIR
 static int path_prefix (const char *s1, const char *s2)
 {
-	return (strncmp (s2, s1, strlen (s1)) == 0);
+	return (strncmp (s2, s1, strlen (s1)) == 0 &&
+		(s2[strlen (s1)] == '\0' || s2[strlen (s1)] == '/'));
 }
 #endif
 
@@ -590,13 +612,10 @@ static void remove_mailbox (void)
  */
 int main (int argc, char **argv)
 {
-	struct passwd *pwd;
-	int arg;
 	int errors = 0;
 
 #ifdef USE_PAM
 	pam_handle_t *pamh = NULL;
-	struct passwd *pampw;
 	int retval;
 #endif
 
@@ -612,16 +631,30 @@ int main (int argc, char **argv)
 	bindtextdomain (PACKAGE, LOCALEDIR);
 	textdomain (PACKAGE);
 
-	while ((arg = getopt (argc, argv, "fr")) != EOF) {
-		switch (arg) {
-		case 'f':	/* force remove even if not owned by user */
-			fflg++;
-			break;
-		case 'r':	/* remove home dir and mailbox */
-			rflg++;
-			break;
-		default:
-			usage ();
+	{
+		/*
+		 * Parse the command line options.
+		 */
+		int c;
+		static struct option long_options[] = {
+			{"force", no_argument, NULL, 'f'},
+			{"help", no_argument, NULL, 'h'},
+			{"remove", no_argument, NULL, 'r'},
+			{NULL, 0, NULL, '\0'}
+		};
+		while ((c =
+			getopt_long (argc, argv, "fhr",
+				     long_options, NULL)) != -1) {
+			switch (c) {
+			case 'f':	/* force remove even if not owned by user */
+				fflg++;
+				break;
+			case 'r':	/* remove home dir and mailbox */
+				rflg++;
+				break;
+			default:
+				usage ();
+			}
 		}
 	}
 
@@ -632,13 +665,19 @@ int main (int argc, char **argv)
 
 #ifdef USE_PAM
 	retval = PAM_SUCCESS;
-	pampw = getpwuid (getuid ());
-	if (pampw == NULL) {
-		retval = PAM_USER_UNKNOWN;
-	}
 
-	if (retval == PAM_SUCCESS)
-		retval = pam_start ("userdel", pampw->pw_name, &conv, &pamh);
+	{
+		struct passwd *pampw;
+		pampw = getpwuid (getuid ()); /* local, no need for xgetpwuid */
+		if (pampw == NULL) {
+			retval = PAM_USER_UNKNOWN;
+		}
+
+		if (retval == PAM_SUCCESS) {
+			retval = pam_start ("userdel", pampw->pw_name,
+					    &conv, &pamh);
+		}
+	}
 
 	if (retval == PAM_SUCCESS) {
 		retval = pam_authenticate (pamh, 0);
@@ -667,14 +706,20 @@ int main (int argc, char **argv)
 	 * Start with a quick check to see if the user exists.
 	 */
 	user_name = argv[argc - 1];
-	if (!(pwd = getpwnam (user_name))) {
-		fprintf (stderr, _("%s: user %s does not exist\n"),
-			 Prog, user_name);
+	{
+		struct passwd *pwd;
+		/* local, no need for xgetpwnam */
+		if (!(pwd = getpwnam (user_name))) {
+			fprintf (stderr, _("%s: user %s does not exist\n"),
+				 Prog, user_name);
 #ifdef WITH_AUDIT
-		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-			      "deleting user not found", user_name, -1, 0);
+			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
+				      "deleting user not found", user_name, -1, 0);
 #endif
-		exit (E_NOTFOUND);
+			exit (E_NOTFOUND);
+		}
+		user_id = pwd->pw_uid;
+		user_home = xstrdup (pwd->pw_dir);
 	}
 #ifdef	USE_NIS
 
@@ -696,8 +741,6 @@ int main (int argc, char **argv)
 		exit (E_NOTFOUND);
 	}
 #endif
-	user_id = pwd->pw_uid;
-	user_home = xstrdup (pwd->pw_dir);
 	/*
 	 * Check to make certain the user isn't logged in.
 	 */
@@ -711,9 +754,6 @@ int main (int argc, char **argv)
 	update_user ();
 	update_groups ();
 
-	nscd_flush_cache ("passwd");
-	nscd_flush_cache ("group");
-
 	if (rflg)
 		remove_mailbox ();
 	if (rflg && !fflg && !is_owner (user_id, user_home)) {
@@ -723,9 +763,11 @@ int main (int argc, char **argv)
 		rflg = 0;
 		errors++;
 	}
+
 #ifdef EXTRA_CHECK_HOME_DIR
 	/* This may be slow, the above should be good enough. */
 	if (rflg && !fflg) {
+		struct passwd *pwd;
 		/*
 		 * For safety, refuse to remove the home directory if it
 		 * would result in removing some other user's home
@@ -752,8 +794,7 @@ int main (int argc, char **argv)
 #endif
 
 	if (rflg) {
-		if (remove_tree (user_home)
-		    || rmdir (user_home)) {
+		if (remove_tree (user_home) != 0) {
 			fprintf (stderr,
 				 _("%s: error removing directory %s\n"),
 				 Prog, user_home);
@@ -776,6 +817,10 @@ int main (int argc, char **argv)
 	 */
 	user_cancel (user_name);
 	close_files ();
+
+	nscd_flush_cache ("passwd");
+	nscd_flush_cache ("group");
+
 #ifdef USE_PAM
 	if (retval == PAM_SUCCESS)
 		pam_end (pamh, PAM_SUCCESS);
