@@ -1,5 +1,8 @@
 /*
- * Copyright 1991 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1991 - 1994, Julianne Frances Haugh
+ * Copyright (c) 1996 - 2000, Marek Michałkiewicz
+ * Copyright (c) 2000 - 2006, Tomasz Kłoczko
+ * Copyright (c) 2007 - 2008, Nicolas François
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -10,26 +13,26 @@
  * 2. Redistributions in binary form must reproduce the above copyright
  *    notice, this list of conditions and the following disclaimer in the
  *    documentation and/or other materials provided with the distribution.
- * 3. Neither the name of Julianne F. Haugh nor the names of its contributors
- *    may be used to endorse or promote products derived from this software
- *    without specific prior written permission.
+ * 3. The name of the copyright holders or contributors may not be used to
+ *    endorse or promote products derived from this software without
+ *    specific prior written permission.
  *
- * THIS SOFTWARE IS PROVIDED BY JULIE HAUGH AND CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED. IN NO EVENT SHALL JULIE HAUGH OR CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
+ * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
+ * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
+ * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
+ * PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT
+ * HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
+ * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
+ * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
+ * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+ * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+ * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
  */
 
 #include <config.h>
 
-#ident "$Id: usermod.c,v 1.69 2006/06/25 13:10:29 kloczek Exp $"
+#ident "$Id: usermod.c 2007 2008-05-24 15:35:15Z nekral-guest $"
 
 #include <ctype.h>
 #include <errno.h>
@@ -68,12 +71,12 @@
 #define E_USAGE		2	/* invalid command syntax */
 #define E_BAD_ARG	3	/* invalid argument to option */
 #define E_UID_IN_USE	4	/* UID already in use (and no -o) */
-/* #define E_BAD_PWFILE     5 *//* passwd file contains errors */
+/* #define E_BAD_PWFILE	5	   passwd file contains errors */
 #define E_NOTFOUND	6	/* specified user/group doesn't exist */
 #define E_USER_BUSY	8	/* user to modify is logged in */
 #define E_NAME_IN_USE	9	/* username already in use */
 #define E_GRP_UPDATE	10	/* can't update group file */
-/* #define E_NOSPACE        11 *//* insufficient space to move home dir */
+/* #define E_NOSPACE	11	   insufficient space to move home dir */
 #define E_HOMEDIR	12	/* unable to complete home dir move */
 #define	VALID(s)	(strcspn (s, ":\n") == strlen (s))
 /*
@@ -87,25 +90,22 @@ static uid_t user_newid;
 static gid_t user_gid;
 static gid_t user_newgid;
 static char *user_comment;
+static char *user_newcomment;
 static char *user_home;
 static char *user_newhome;
 static char *user_shell;
+static char *user_newshell;
 static long user_expire;
+static long user_newexpire;
 static long user_inactive;
+static long user_newinactive;
 static long sys_ngroups;
 static char **user_groups;	/* NULL-terminated list */
 
-#ifdef WITH_AUDIT
-static char *user_newcomment;	/* Audit */
-static char *user_newshell;	/* Audit */
-static long user_newexpire;	/* Audit */
-static long user_newinactive;	/* Audit */
-#endif
-
 static char *Prog;
 
 static int
- aflg = 0,			/* append to existing secondary group set */
+    aflg = 0,			/* append to existing secondary group set */
     cflg = 0,			/* new comment (GECOS) field */
     dflg = 0,			/* new home directory */
     eflg = 0,			/* days since 1970-01-01 when account becomes expired */
@@ -127,20 +127,29 @@ static int is_shadow_pwd;
 static int is_shadow_grp;
 #endif
 
+static int pw_locked  = 0;
+static int spw_locked = 0;
+static int gr_locked  = 0;
+#ifdef SHADOWGRP
+static int sgr_locked = 0;
+#endif
+
 
 /* local function prototypes */
+static void date_to_str (char *buf, size_t maxsize,
+                         long int date, const char *negativ);
 static int get_groups (char *);
 static void usage (void);
 static void new_pwent (struct passwd *);
 
 static void new_spent (struct spwd *);
 static void fail_exit (int);
-static int update_group (void);
+static void update_group (void);
 
 #ifdef SHADOWGRP
-static int update_gshadow (void);
+static void update_gshadow (void);
 #endif
-static int grp_update (void);
+static void grp_update (void);
 
 static long get_number (const char *);
 static uid_t get_id (const char *);
@@ -155,21 +164,39 @@ static void update_files (void);
 static void move_mailbox (void);
 #endif
 
+static void date_to_str (char *buf, size_t maxsize,
+                         long int date, const char *negativ)
+{
+	struct tm *tp;
+
+	if ((negativ != NULL) && (date < 0)) {
+		strncpy (buf, negativ, maxsize);
+	} else {
+		time_t t = date;
+		tp = gmtime (&t);
+#ifdef HAVE_STRFTIME
+		strftime (buf, maxsize, "%Y-%m-%d", tp);
+#else
+		snprintf (buf, maxsize, "%04d-%02d-%02d",
+		          tp->tm_year + 1900, tp->tm_mon + 1, tp->tm_mday);
+#endif				/* HAVE_STRFTIME */
+	}
+	buf[maxsize - 1] = '\0';
+}
 /*
  * Had to move this over from useradd.c since we have groups named
  * "56k-family"... ergh.
  * --Pac.
  */
-static struct group *getgr_nam_gid (const char *name)
+static struct group *getgr_nam_gid (const char *grname)
 {
-	gid_t gid;
-	char *ep;
+	long val;
+	char *errptr;
 
-	gid = strtoul (name, &ep, 10);
-	if (*name != '\0' && *ep == '\0')	/* valid numeric GID */
-		return getgrgid (gid);
-
-	return getgrnam (name);
+	val = strtol (grname, &errptr, 10);
+	if (*grname != '\0' && *errptr == '\0' && errno != ERANGE && val >= 0)
+		return xgetgrgid (val);
+	return xgetgrnam (grname);
 }
 
 /*
@@ -273,28 +300,30 @@ static int get_groups (char *list)
  */
 static void usage (void)
 {
-	fprintf (stderr, _("Usage: usermod [options] LOGIN\n"
-			   "\n"
-			   "Options:\n"
-			   "  -a, --append GROUP		append the user to the supplemental GROUP\n"
-			   "  -c, --comment COMMENT		new value of the GECOS field\n"
-			   "  -d, --home HOME_DIR		new home directory for the user account\n"
-			   "  -e, --expiredate EXPIRE_DATE	set account expiration date to EXPIRE_DATE\n"
-			   "  -f, --inactive INACTIVE	set password inactive after expiration\n"
-			   "				to INACTIVE\n"
-			   "  -g, --gid GROUP		force use GROUP as new primary group\n"
-			   "  -G, --groups GROUPS		new list of supplementary GROUPS\n"
-			   "  -h, --help			display this help message and exit\n"
-			   "  -l, --login NEW_LOGIN		new value of the login name\n"
-			   "  -L, --lock			lock the user account\n"
-			   "  -m, --move-home		move contents of the home directory to the new\n"
-			   "				location (use only with -d)\n"
-			   "  -o, --non-unique		allow using duplicate (non-unique) UID\n"
-			   "  -p, --password PASSWORD	use encrypted password for the new password\n"
-			   "  -s, --shell SHELL		new login shell for the user account\n"
-			   "  -u, --uid UID			new UID for the user account\n"
-			   "  -U, --unlock			unlock the user account\n"
-			   "\n"));
+	fputs (_("Usage: usermod [options] LOGIN\n"
+	         "\n"
+	         "Options:\n"
+	         "  -c, --comment COMMENT         new value of the GECOS field\n"
+	         "  -d, --home HOME_DIR           new home directory for the user account\n"
+	         "  -e, --expiredate EXPIRE_DATE  set account expiration date to EXPIRE_DATE\n"
+	         "  -f, --inactive INACTIVE       set password inactive after expiration\n"
+	         "                                to INACTIVE\n"
+	         "  -g, --gid GROUP               force use GROUP as new primary group\n"
+	         "  -G, --groups GROUPS           new list of supplementary GROUPS\n"
+	         "  -a, --append                  append the user to the supplemental GROUPS\n"
+	         "                                mentioned by the -G option without removing\n"
+	         "                                him/her from other groups\n"
+	         "  -h, --help                    display this help message and exit\n"
+	         "  -l, --login NEW_LOGIN         new value of the login name\n"
+	         "  -L, --lock                    lock the user account\n"
+	         "  -m, --move-home               move contents of the home directory to the\n"
+	         "                                new location (use only with -d)\n"
+	         "  -o, --non-unique              allow using duplicate (non-unique) UID\n"
+	         "  -p, --password PASSWORD       use encrypted password for the new password\n"
+	         "  -s, --shell SHELL             new login shell for the user account\n"
+	         "  -u, --uid UID                 new UID for the user account\n"
+	         "  -U, --unlock                  unlock the user account\n"
+	         "\n"), stderr);
 	exit (E_USAGE);
 }
 
@@ -302,7 +331,7 @@ static void usage (void)
  * update encrypted password string (for both shadow and non-shadow
  * passwords)
  */
-static char *new_pw_passwd (char *pw_pass, const char *pw_name)
+static char *new_pw_passwd (char *pw_pass)
 {
 	if (Lflg && pw_pass[0] != '!') {
 		char *buf = xmalloc (strlen (pw_pass) + 2);
@@ -311,18 +340,26 @@ static char *new_pw_passwd (char *pw_pass, const char *pw_name)
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "updating passwd",
 			      user_newname, user_newid, 0);
 #endif
-		SYSLOG ((LOG_INFO, "lock user `%s' password", pw_name));
+		SYSLOG ((LOG_INFO, "lock user `%s' password", user_newname));
 		strcpy (buf, "!");
 		strcat (buf, pw_pass);
 		pw_pass = buf;
 	} else if (Uflg && pw_pass[0] == '!') {
 		char *s;
 
+		if (pw_pass[1] == '\0') {
+			fprintf (stderr,
+				 _("%s: unlocking the user would result in a passwordless account.\n"
+				   "You should set a password with usermod -p to unlock this user account.\n"),
+				 Prog);
+			return pw_pass;
+		}
+
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "updating password",
 			      user_newname, user_newid, 0);
 #endif
-		SYSLOG ((LOG_INFO, "unlock user `%s' password", pw_name));
+		SYSLOG ((LOG_INFO, "unlock user `%s' password", user_newname));
 		s = pw_pass;
 		while (*s) {
 			*s = *(s + 1);
@@ -333,7 +370,7 @@ static char *new_pw_passwd (char *pw_pass, const char *pw_name)
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "changing password",
 			      user_newname, user_newid, 1);
 #endif
-		SYSLOG ((LOG_INFO, "change user `%s' password", pw_name));
+		SYSLOG ((LOG_INFO, "change user `%s' password", user_newname));
 		pw_pass = xstrdup (user_pass);
 	}
 	return pw_pass;
@@ -358,7 +395,7 @@ static void new_pwent (struct passwd *pwent)
 	}
 	if (!is_shadow_pwd)
 		pwent->pw_passwd =
-		    new_pw_passwd (pwent->pw_passwd, pwent->pw_name);
+		    new_pw_passwd (pwent->pw_passwd);
 
 	if (uflg) {
 #ifdef WITH_AUDIT
@@ -385,10 +422,8 @@ static void new_pwent (struct passwd *pwent)
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "changing comment",
 			      user_newname, user_newid, 1);
-		pwent->pw_gecos = user_newcomment;
-#else
-		pwent->pw_gecos = user_comment;
 #endif
+		pwent->pw_gecos = user_newcomment;
 	}
 
 	if (dflg) {
@@ -406,15 +441,10 @@ static void new_pwent (struct passwd *pwent)
 #ifdef WITH_AUDIT
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "changing user shell",
 			      user_newname, user_newid, 1);
+#endif
 		SYSLOG ((LOG_INFO, "change user `%s' shell from `%s' to `%s'",
 			 pwent->pw_name, pwent->pw_shell, user_newshell));
 		pwent->pw_shell = user_newshell;
-#else
-		SYSLOG ((LOG_INFO,
-			 "change user `%s' shell from `%s' to `%s'",
-			 pwent->pw_name, pwent->pw_shell, user_shell));
-		pwent->pw_shell = user_shell;
-#endif
 	}
 }
 
@@ -434,76 +464,30 @@ static void new_spent (struct spwd *spent)
 		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
 			      "changing inactive days", user_newname,
 			      user_newid, 1);
+#endif
 		SYSLOG ((LOG_INFO,
 			 "change user `%s' inactive from `%ld' to `%ld'",
 			 spent->sp_namp, spent->sp_inact, user_newinactive));
 		spent->sp_inact = user_newinactive;
-#else
-
-		SYSLOG ((LOG_INFO,
-			 "change user `%s' inactive from `%ld' to `%ld'",
-			 spent->sp_namp, spent->sp_inact, user_inactive));
-		spent->sp_inact = user_inactive;
-#endif
 	}
 	if (eflg) {
-		/* XXX - dates might be better than numbers of days.  --marekm */
+		/* log dates rather than numbers of days. */
+		char new_exp[16], old_exp[16];
+		date_to_str (new_exp, sizeof(new_exp),
+		             user_newexpire * DAY, "never");
+		date_to_str (old_exp, sizeof(old_exp),
+		             user_expire * DAY, "never");
 #ifdef WITH_AUDIT
-		if (audit_fd >= 0) {
-			time_t exp_t;
-			struct tm *exp_tm;
-			char new_exp[16], old_exp[16];
-
-			if (user_newexpire == -1)
-				new_exp[0] = '\0';
-			else {
-				exp_t = user_newexpire * DAY;
-				exp_tm = gmtime (&exp_t);
-#ifdef HAVE_STRFTIME
-				strftime (new_exp, sizeof (new_exp), "%Y-%m-%d",
-					  exp_tm);
-#else
-				memset (new_exp, 0, sizeof (new_exp));
-				snprintf (new_exp, sizeof (new_exp) - 1,
-					  "%04i-%02i-%02i",
-					  exp_tm->tm_year + 1900,
-					  exp_tm->tm_mon + 1, exp_tm->tm_mday);
+		audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
+		              "changing expiration date", user_newname,
+		              user_newid, 1);
 #endif
-			}
-
-			if (user_expire == -1)
-				old_exp[0] = '\0';
-			else {
-				exp_t = user_expire * DAY;
-				exp_tm = gmtime (&exp_t);
-#ifdef HAVE_STRFTIME
-				strftime (old_exp, sizeof (old_exp), "%Y-%m-%d",
-					  exp_tm);
-#else
-				memset (old_exp, 0, sizeof (old_exp));
-				snprintf (old_exp, sizeof (old_exp) - 1,
-					  "%04i-%02i-%02i",
-					  exp_tm->tm_year + 1900,
-					  exp_tm->tm_mon + 1, exp_tm->tm_mday);
-#endif
-			}
-			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
-				      "changing expiration date", user_newname,
-				      user_newid, 1);
-		}
-
 		SYSLOG ((LOG_INFO,
-			 "change user `%s' expiration from `%ld' to `%ld'",
-			 spent->sp_namp, spent->sp_expire, user_newexpire));
+			 "change user `%s' expiration from `%s' to `%s'",
+			 spent->sp_namp, old_exp, new_exp));
 		spent->sp_expire = user_newexpire;
-#else
-		SYSLOG ((LOG_INFO,
-			 "change user `%s' expiration from `%ld' to `%ld'",
-			 spent->sp_namp, spent->sp_expire, user_expire));
-		spent->sp_expire = user_expire;
-#endif
 	}
-	spent->sp_pwdp = new_pw_passwd (spent->sp_pwdp, spent->sp_namp);
+	spent->sp_pwdp = new_pw_passwd (spent->sp_pwdp);
 	if (pflg)
 		spent->sp_lstchg = time ((time_t *) 0) / SCALE;
 }
@@ -513,14 +497,17 @@ static void new_spent (struct spwd *spent)
  */
 static void fail_exit (int code)
 {
-	(void) gr_unlock ();
+	if (gr_locked)
+		gr_unlock ();
 #ifdef	SHADOWGRP
-	if (is_shadow_grp)
+	if (sgr_locked)
 		sgr_unlock ();
 #endif
-	if (is_shadow_pwd)
+	if (spw_locked)
 		spw_unlock ();
-	(void) pw_unlock ();
+	if (pw_locked)
+		pw_unlock ();
+
 #ifdef WITH_AUDIT
 	audit_logger (AUDIT_USER_CHAUTHTOK, Prog, "modifying account",
 		      user_name, -1, 0);
@@ -529,7 +516,7 @@ static void fail_exit (int code)
 }
 
 
-static int update_group (void)
+static void update_group (void)
 {
 	int is_member;
 	int was_member;
@@ -537,22 +524,6 @@ static int update_group (void)
 	const struct group *grp;
 	struct group *ngrp;
 
-	/*
-	 * Lock and open the group file. This will load all of the group
-	 * entries.
-	 */
-	if (!gr_lock ()) {
-		fprintf (stderr, _("%s: error locking group file\n"), Prog);
-		SYSLOG ((LOG_ERR, "error locking group file"));
-		return -1;
-	}
-	if (!gr_open (O_RDWR)) {
-		fprintf (stderr, _("%s: error opening group file\n"), Prog);
-		SYSLOG ((LOG_ERR, "error opening group file"));
-		gr_unlock ();
-		return -1;
-	}
-
 	changed = 0;
 
 	/*
@@ -565,7 +536,8 @@ static int update_group (void)
 		 * concurrent groups.
 		 */
 		was_member = is_on_list (grp->gr_mem, user_name);
-		is_member = Gflg && is_on_list (user_groups, grp->gr_name);
+		is_member = Gflg && (   (was_member && aflg)
+		                     || is_on_list (user_groups, grp->gr_name));
 
 		if (!was_member && !is_member)
 			continue;
@@ -573,10 +545,9 @@ static int update_group (void)
 		ngrp = __gr_dup (grp);
 		if (!ngrp) {
 			fprintf (stderr,
-				 _("%s: out of memory in update_group\n"),
+				 _("%s: Out of memory. Cannot update the group database.\n"),
 				 Prog);
-			gr_unlock ();
-			return -1;
+			fail_exit (E_GRP_UPDATE);
 		}
 
 		if (was_member && (!Gflg || is_member)) {
@@ -607,17 +578,14 @@ static int update_group (void)
 			SYSLOG ((LOG_INFO, "delete `%s' from group `%s'",
 				 user_name, ngrp->gr_name));
 		} else if (!was_member && Gflg && is_member) {
-			ngrp->gr_mem = add_list (ngrp->gr_mem,
-						 lflg ? user_newname :
-						 user_name);
+			ngrp->gr_mem = add_list (ngrp->gr_mem, user_newname);
 			changed = 1;
 #ifdef WITH_AUDIT
 			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
 				      "adding user to group", user_name, -1, 1);
 #endif
 			SYSLOG ((LOG_INFO, "add `%s' to group `%s'",
-				 lflg ? user_newname : user_name,
-				 ngrp->gr_name));
+			         user_newname, ngrp->gr_name));
 		}
 		if (!changed)
 			continue;
@@ -626,23 +594,14 @@ static int update_group (void)
 		if (!gr_update (ngrp)) {
 			fprintf (stderr,
 				 _("%s: error adding new group entry\n"), Prog);
-			SYSLOG ((LOG_ERR, "error adding group entry"));
-			gr_unlock ();
-			return -1;
+			SYSLOG ((LOG_ERR, "error adding new group entry"));
+			fail_exit (E_GRP_UPDATE);
 		}
 	}
-
-	if (!gr_close ()) {
-		fprintf (stderr, _("%s: cannot rewrite group file\n"), Prog);
-		gr_unlock ();
-		return -1;
-	}
-	gr_unlock ();
-	return 0;
 }
 
 #ifdef SHADOWGRP
-static int update_gshadow (void)
+static void update_gshadow (void)
 {
 	int is_member;
 	int was_member;
@@ -651,20 +610,6 @@ static int update_gshadow (void)
 	const struct sgrp *sgrp;
 	struct sgrp *nsgrp;
 
-	if (!sgr_lock ()) {
-		fprintf (stderr,
-			 _("%s: error locking shadow group file\n"), Prog);
-		SYSLOG ((LOG_ERR, "error locking shadow group file"));
-		return -1;
-	}
-	if (!sgr_open (O_RDWR)) {
-		fprintf (stderr,
-			 _("%s: error opening shadow group file\n"), Prog);
-		SYSLOG ((LOG_ERR, "error opening shadow group file"));
-		sgr_unlock ();
-		return -1;
-	}
-
 	changed = 0;
 
 	/*
@@ -688,6 +633,8 @@ static int update_gshadow (void)
 		 * concurrent groups.
 		 */
 		is_member = Gflg && is_on_list (user_groups, sgrp->sg_name);
+		is_member = Gflg && (   (was_member && aflg)
+		                     || is_on_list (user_groups, sgrp->sg_name));
 
 		if (!was_member && !was_admin && !is_member)
 			continue;
@@ -695,11 +642,9 @@ static int update_gshadow (void)
 		nsgrp = __sgr_dup (sgrp);
 		if (!nsgrp) {
 			fprintf (stderr,
-				 _
-				 ("%s: out of memory in update_gshadow\n"),
+				 _("%s: Out of memory. Cannot update the shadow group database.\n"),
 				 Prog);
-			sgr_unlock ();
-			return -1;
+			fail_exit (E_GRP_UPDATE);
 		}
 
 		if (was_admin && lflg) {
@@ -744,9 +689,7 @@ static int update_gshadow (void)
 				 "delete `%s' from shadow group `%s'",
 				 user_name, nsgrp->sg_name));
 		} else if (!was_member && Gflg && is_member) {
-			nsgrp->sg_mem = add_list (nsgrp->sg_mem,
-						  lflg ? user_newname :
-						  user_name);
+			nsgrp->sg_mem = add_list (nsgrp->sg_mem, user_newname);
 			changed = 1;
 #ifdef WITH_AUDIT
 			audit_logger (AUDIT_USER_CHAUTHTOK, Prog,
@@ -754,8 +697,7 @@ static int update_gshadow (void)
 				      user_newname, -1, 1);
 #endif
 			SYSLOG ((LOG_INFO, "add `%s' to shadow group `%s'",
-				 lflg ? user_newname : user_name,
-				 nsgrp->sg_name));
+			         user_newname, nsgrp->sg_name));
 		}
 		if (!changed)
 			continue;
@@ -767,21 +709,11 @@ static int update_gshadow (void)
 		 */
 		if (!sgr_update (nsgrp)) {
 			fprintf (stderr,
-				 _("%s: error adding new group entry\n"), Prog);
+				 _("%s: error adding new shadow group entry\n"), Prog);
 			SYSLOG ((LOG_ERR, "error adding shadow group entry"));
-			sgr_unlock ();
-			return -1;
+			fail_exit (E_GRP_UPDATE);
 		}
 	}
-
-	if (!sgr_close ()) {
-		fprintf (stderr,
-			 _("%s: cannot rewrite shadow group file\n"), Prog);
-		sgr_unlock ();
-		return -1;
-	}
-	sgr_unlock ();
-	return 0;
 }
 #endif				/* SHADOWGRP */
 
@@ -791,42 +723,41 @@ static int update_gshadow (void)
  *	grp_update() takes the secondary group set given in user_groups and
  *	adds the user to each group given by that set.
  */
-static int grp_update (void)
+static void grp_update (void)
 {
-	int ret;
-
-	ret = update_group ();
+	update_group ();
 #ifdef SHADOWGRP
-	if (!ret && is_shadow_grp)
-		ret = update_gshadow ();
+	if (is_shadow_grp)
+		update_gshadow ();
 #endif
-	return ret;
 }
 
-static long get_number (const char *cp)
+static long get_number (const char *numstr)
 {
 	long val;
-	char *ep;
+	char *errptr;
 
-	val = strtol (cp, &ep, 10);
-	if (*cp != '\0' && *ep == '\0')	/* valid number */
-		return val;
-
-	fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog, cp);
-	exit (E_BAD_ARG);
+	val = strtol (numstr, &errptr, 10);
+	if (*errptr || errno == ERANGE) {
+		fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog,
+			 numstr);
+		exit (E_BAD_ARG);
+	}
+	return val;
 }
 
-static uid_t get_id (const char *cp)
+static uid_t get_id (const char *uidstr)
 {
-	uid_t val;
-	char *ep;
+	long val;
+	char *errptr;
 
-	val = strtoul (cp, &ep, 10);
-	if (*cp != '\0' && *ep == '\0')	/* valid number */
-		return val;
-
-	fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog, cp);
-	exit (E_BAD_ARG);
+	val = strtol (uidstr, &errptr, 10);
+	if (*errptr || errno == ERANGE || val < 0) {
+		fprintf (stderr, _("%s: invalid numeric argument '%s'\n"), Prog,
+			 uidstr);
+		exit (E_BAD_ARG);
+	}
+	return val;
 }
 
 /*
@@ -839,34 +770,34 @@ static uid_t get_id (const char *cp)
 static void process_flags (int argc, char **argv)
 {
 	const struct group *grp;
-	const struct passwd *pwd;
 
-	const struct spwd *spwd = NULL;
 	int anyflag = 0;
 
 	if (argc == 1 || argv[argc - 1][0] == '-')
 		usage ();
 
-	if (!(pwd = getpwnam (argv[argc - 1]))) {
-		fprintf (stderr, _("%s: user %s does not exist\n"),
-			 Prog, argv[argc - 1]);
-		exit (E_NOTFOUND);
-	}
+	{
+		const struct passwd *pwd;
+		/* local, no need for xgetpwnam */
+		if (!(pwd = getpwnam (argv[argc - 1]))) {
+			fprintf (stderr, _("%s: user %s does not exist\n"),
+				 Prog, argv[argc - 1]);
+			exit (E_NOTFOUND);
+		}
 
-	user_name = argv[argc - 1];
-	user_id = pwd->pw_uid;
-	user_gid = pwd->pw_gid;
-	user_comment = xstrdup (pwd->pw_gecos);
-	user_home = xstrdup (pwd->pw_dir);
-	user_shell = xstrdup (pwd->pw_shell);
-#ifdef WITH_AUDIT
+		user_name = argv[argc - 1];
+		user_id = pwd->pw_uid;
+		user_gid = pwd->pw_gid;
+		user_comment = xstrdup (pwd->pw_gecos);
+		user_home = xstrdup (pwd->pw_dir);
+		user_shell = xstrdup (pwd->pw_shell);
+	}
 	user_newname = user_name;
 	user_newid = user_id;
 	user_newgid = user_gid;
 	user_newcomment = user_comment;
 	user_newhome = user_home;
 	user_newshell = user_shell;
-#endif
 
 #ifdef	USE_NIS
 	/*
@@ -888,13 +819,15 @@ static void process_flags (int argc, char **argv)
 	}
 #endif
 
-	if (is_shadow_pwd && (spwd = getspnam (user_name))) {
-		user_expire = spwd->sp_expire;
-		user_inactive = spwd->sp_inact;
-#ifdef WITH_AUDIT
-		user_newexpire = user_expire;
-		user_newinactive = user_inactive;
-#endif
+	{
+		const struct spwd *spwd = NULL;
+		/* local, no need for xgetspnam */
+		if (is_shadow_pwd && (spwd = getspnam (user_name))) {
+			user_expire = spwd->sp_expire;
+			user_inactive = spwd->sp_inact;
+			user_newexpire = user_expire;
+			user_newinactive = user_inactive;
+		}
 	}
 
 	{
@@ -903,7 +836,7 @@ static void process_flags (int argc, char **argv)
 		 */
 		int c;
 		static struct option long_options[] = {
-			{"append", required_argument, NULL, 'a'},
+			{"append", no_argument, NULL, 'a'},
 			{"comment", required_argument, NULL, 'c'},
 			{"home", required_argument, NULL, 'd'},
 			{"expiredate", required_argument, NULL, 'e'},
@@ -922,7 +855,7 @@ static void process_flags (int argc, char **argv)
 			{NULL, 0, NULL, '\0'}
 		};
 		while ((c =
-			getopt_long (argc, argv, "a:c:d:e:f:g:G:hl:Lmop:s:u:U",
+			getopt_long (argc, argv, "ac:d:e:f:g:G:hl:Lmop:s:u:U",
 				     long_options, NULL)) != -1) {
 			switch (c) {
 			case 'a':
@@ -935,11 +868,7 @@ static void process_flags (int argc, char **argv)
 						 Prog, optarg);
 					exit (E_BAD_ARG);
 				}
-#ifdef WITH_AUDIT
 				user_newcomment = optarg;
-#else
-				user_comment = optarg;
-#endif
 				cflg++;
 				break;
 			case 'd':
@@ -954,38 +883,21 @@ static void process_flags (int argc, char **argv)
 				break;
 			case 'e':
 				if (*optarg) {
-#ifdef WITH_AUDIT
 					user_newexpire = strtoday (optarg);
 					if (user_newexpire == -1) {
-#else
-					user_expire = strtoday (optarg);
-					if (user_expire == -1) {
-#endif
 						fprintf (stderr,
 							 _
 							 ("%s: invalid date '%s'\n"),
 							 Prog, optarg);
 						exit (E_BAD_ARG);
 					}
-#ifdef WITH_AUDIT
 					user_newexpire *= DAY / SCALE;
-#else
-					user_expire *= DAY / SCALE;
-#endif
 				} else
-#ifdef WITH_AUDIT
 					user_newexpire = -1;
-#else
-					user_expire = -1;
-#endif
 				eflg++;
 				break;
 			case 'f':
-#ifdef WITH_AUDIT
 				user_newinactive = get_number (optarg);
-#else
-				user_inactive = get_number (optarg);
-#endif
 				fflg++;
 				break;
 			case 'g':
@@ -1011,39 +923,19 @@ static void process_flags (int argc, char **argv)
 						 Prog, optarg);
 					exit (E_BAD_ARG);
 				}
-
-				/*
-				 * If the name does not really change, we mustn't
-				 * set the flag as this will cause rather serious
-				 * problems later!
-				 */
-				if (strcmp (user_name, optarg))
-					lflg++;
-
+				lflg++;
 				user_newname = optarg;
 				break;
 			case 'L':
-				if (Uflg || pflg)
-					usage ();
-
 				Lflg++;
 				break;
 			case 'm':
-				if (!dflg)
-					usage ();
-
 				mflg++;
 				break;
 			case 'o':
-				if (!uflg)
-					usage ();
-
 				oflg++;
 				break;
 			case 'p':
-				if (Lflg || Uflg)
-					usage ();
-
 				user_pass = optarg;
 				pflg++;
 				break;
@@ -1054,11 +946,7 @@ static void process_flags (int argc, char **argv)
 						 Prog, optarg);
 					exit (E_BAD_ARG);
 				}
-#ifdef WITH_AUDIT
 				user_newshell = optarg;
-#else
-				user_shell = optarg;
-#endif
 				sflg++;
 				break;
 			case 'u':
@@ -1066,9 +954,6 @@ static void process_flags (int argc, char **argv)
 				uflg++;
 				break;
 			case 'U':
-				if (Lflg && pflg)
-					usage ();
-
 				Uflg++;
 				break;
 			default:
@@ -1082,6 +967,40 @@ static void process_flags (int argc, char **argv)
 		fprintf (stderr, _("%s: no flags given\n"), Prog);
 		exit (E_USAGE);
 	}
+
+	if (user_newid == user_id) {
+		uflg = 0;
+		oflg = 0;
+	}
+	if (user_newgid == user_gid) {
+		gflg = 0;
+	}
+	if (strcmp (user_newshell, user_shell) == 0) {
+		sflg = 0;
+	}
+	if (strcmp (user_newname, user_name) == 0) {
+		lflg = 0;
+	}
+	if (user_newinactive == user_inactive) {
+		fflg = 0;
+	}
+	if (user_newexpire == user_expire) {
+		eflg = 0;
+	}
+	if (strcmp (user_newhome, user_home) == 0) {
+		dflg = 0;
+		mflg = 0;
+	}
+	if (strcmp (user_newcomment, user_comment) == 0) {
+		cflg = 0;
+	}
+
+	if (Uflg + uflg + sflg + pflg + oflg + mflg + Lflg + lflg + Gflg +
+	    gflg + fflg + eflg + dflg + cflg == 0) {
+		fprintf (stderr, _("%s: no changes\n"), Prog);
+		exit (E_SUCCESS);
+	}
+
 	if (!is_shadow_pwd && (eflg || fflg)) {
 		fprintf (stderr,
 			 _
@@ -1095,23 +1014,43 @@ static void process_flags (int argc, char **argv)
 
 	if (aflg && (!Gflg)) {
 		fprintf (stderr,
-			 _("%s: -a flag is ONLY allowed with the -G flag\n"),
+			 _("%s: %s flag is ONLY allowed with the %s flag\n"),
+			 Prog, "-a", "-G");
+		usage ();
+		exit (E_USAGE);
+	}
+
+	if ((Lflg && (pflg || Uflg)) || (pflg && Uflg)) {
+		fprintf (stderr,
+			 _("%s: the -L, -p, and -U flags are exclusive\n"),
 			 Prog);
 		usage ();
 		exit (E_USAGE);
 	}
 
-	if (dflg && strcmp (user_home, user_newhome) == 0)
-		dflg = mflg = 0;
+	if (oflg && !uflg) {
+		fprintf (stderr,
+			 _("%s: %s flag is ONLY allowed with the %s flag\n"),
+			 Prog, "-o", "-u");
+		usage ();
+		exit (E_USAGE);
+	}
 
-	if (uflg && user_id == user_newid)
-		uflg = oflg = 0;
+	if (mflg && !dflg) {
+		fprintf (stderr,
+			 _("%s: %s flag is ONLY allowed with the %s flag\n"),
+			 Prog, "-m", "-d");
+		usage ();
+		exit (E_USAGE);
+	}
 
+	/* local, no need for xgetpwnam */
 	if (lflg && getpwnam (user_newname)) {
 		fprintf (stderr, _("%s: user %s exists\n"), Prog, user_newname);
 		exit (E_NAME_IN_USE);
 	}
 
+	/* local, no need for xgetpwuid */
 	if (uflg && !oflg && getpwuid (user_newid)) {
 		fprintf (stderr, _("%s: uid %lu is not unique\n"),
 			 Prog, (unsigned long) user_newid);
@@ -1136,9 +1075,36 @@ static void close_files (void)
 			 _("%s: cannot rewrite shadow password file\n"), Prog);
 		fail_exit (E_PW_UPDATE);
 	}
+
+	if (Gflg || lflg) {
+		if (!gr_close ()) {
+			fprintf (stderr, _("%s: cannot rewrite group file\n"),
+				 Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+#ifdef SHADOWGRP
+		if (is_shadow_grp && !sgr_close ()) {
+			fprintf (stderr,
+				 _("%s: cannot rewrite shadow group file\n"),
+				 Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+		if (is_shadow_grp)
+			sgr_unlock ();
+#endif
+		gr_unlock ();
+	}
+
 	if (is_shadow_pwd)
 		spw_unlock ();
-	(void) pw_unlock ();
+	pw_unlock ();
+
+	pw_locked = 0;
+	spw_locked = 0;
+	gr_locked = 0;
+#ifdef	SHADOWGRP
+	sgr_locked = 0;
+#endif
 
 	/*
 	 * Close the DBM and/or flat files
@@ -1160,8 +1126,9 @@ static void open_files (void)
 {
 	if (!pw_lock ()) {
 		fprintf (stderr, _("%s: unable to lock password file\n"), Prog);
-		exit (E_PW_UPDATE);
+		fail_exit (E_PW_UPDATE);
 	}
+	pw_locked = 1;
 	if (!pw_open (O_RDWR)) {
 		fprintf (stderr, _("%s: unable to open password file\n"), Prog);
 		fail_exit (E_PW_UPDATE);
@@ -1171,11 +1138,45 @@ static void open_files (void)
 			 _("%s: cannot lock shadow password file\n"), Prog);
 		fail_exit (E_PW_UPDATE);
 	}
+	spw_locked = 1;
 	if (is_shadow_pwd && !spw_open (O_RDWR)) {
 		fprintf (stderr,
 			 _("%s: cannot open shadow password file\n"), Prog);
 		fail_exit (E_PW_UPDATE);
 	}
+
+	if (Gflg || lflg) {
+		/*
+		 * Lock and open the group file. This will load all of the
+		 * group entries.
+		 */
+		if (!gr_lock ()) {
+			fprintf (stderr, _("%s: error locking group file\n"),
+				 Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+		gr_locked = 1;
+		if (!gr_open (O_RDWR)) {
+			fprintf (stderr, _("%s: error opening group file\n"),
+				 Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+#ifdef SHADOWGRP
+		if (is_shadow_grp && !sgr_lock ()) {
+			fprintf (stderr,
+				 _("%s: error locking shadow group file\n"),
+				 Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+		sgr_locked = 1;
+		if (is_shadow_grp && !sgr_open (O_RDWR)) {
+			fprintf (stderr,
+				 _("%s: error opening shadow group file\n"),
+				 Prog);
+			fail_exit (E_GRP_UPDATE);
+		}
+#endif
+	}
 }
 
 /*
@@ -1285,14 +1286,14 @@ static void move_home (void)
 					fail_exit (E_HOMEDIR);
 				}
 				if (copy_tree (user_home, user_newhome,
-					       uflg ? user_newid : -1,
-					       gflg ? user_newgid : -1) == 0) {
-					if (remove_tree (user_home) != 0 ||
-					    rmdir (user_home) != 0)
+					       uflg ? (long int)user_newid : -1,
+					       gflg ? (long int)user_newgid : -1) == 0) {
+					if (remove_tree (user_home) != 0) {
 						fprintf (stderr,
 							 _
 							 ("%s: warning: failed to completely remove old home directory %s"),
 							 Prog, user_home);
+					}
 #ifdef WITH_AUDIT
 					audit_logger (AUDIT_USER_CHAUTHTOK,
 						      Prog,
@@ -1303,8 +1304,9 @@ static void move_home (void)
 					return;
 				}
 
+				/* TODO: do some cleanup if the copy
+				 *       was started */
 				(void) remove_tree (user_newhome);
-				(void) rmdir (user_newhome);
 			}
 			fprintf (stderr,
 				 _
@@ -1452,11 +1454,8 @@ static void move_mailbox (void)
  */
 int main (int argc, char **argv)
 {
-	int grp_err = 0;
-
 #ifdef USE_PAM
 	pam_handle_t *pamh = NULL;
-	struct passwd *pampw;
 	int retval;
 #endif
 
@@ -1489,13 +1488,17 @@ int main (int argc, char **argv)
 #ifdef USE_PAM
 	retval = PAM_SUCCESS;
 
-	pampw = getpwuid (getuid ());
-	if (pampw == NULL) {
-		retval = PAM_USER_UNKNOWN;
-	}
+	{
+		struct passwd *pampw;
+		pampw = getpwuid (getuid ()); /* local, no need for xgetpwuid */
+		if (pampw == NULL) {
+			retval = PAM_USER_UNKNOWN;
+		}
 
-	if (retval == PAM_SUCCESS) {
-		retval = pam_start ("usermod", pampw->pw_name, &conv, &pamh);
+		if (retval == PAM_SUCCESS) {
+			retval = pam_start ("usermod", pampw->pw_name,
+					    &conv, &pamh);
+		}
 	}
 
 	if (retval == PAM_SUCCESS) {
@@ -1523,15 +1526,13 @@ int main (int argc, char **argv)
 	 * change the home directory, then close and update the files.
 	 */
 	open_files ();
-
 	usr_update ();
-	nscd_flush_cache ("passwd");
-	nscd_flush_cache ("group");
-
+	if (Gflg || lflg)
+		grp_update ();
 	close_files ();
 
-	if (Gflg || lflg)
-		grp_err = grp_update ();
+	nscd_flush_cache ("passwd");
+	nscd_flush_cache ("group");
 
 	if (mflg)
 		move_home ();
@@ -1553,9 +1554,6 @@ int main (int argc, char **argv)
 			    user_gid, gflg ? user_newgid : user_gid);
 	}
 
-	if (grp_err)
-		exit (E_GRP_UPDATE);
-
 #ifdef USE_PAM
 	if (retval == PAM_SUCCESS)
 		pam_end (pamh, PAM_SUCCESS);
